#!/usr/bin/env python
# cardinal_pythonlib/wsgi/reverse_proxied_mw.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Middleware to set SCRIPT_NAME environment variable etc. when behind a
reverse proxy.**

"""

from pprint import pformat
from typing import List

from cardinal_pythonlib.dicts import dict_diff, delete_keys
from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler
from cardinal_pythonlib.wsgi.constants import (
    TYPE_WSGI_APP,
    TYPE_WSGI_APP_RESULT,
    TYPE_WSGI_ENVIRON,
    TYPE_WSGI_START_RESPONSE,
    WsgiEnvVar,
)

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Helper functions for handling HTTP headers
# =============================================================================

def ip_addresses_from_xff(value: str) -> List[str]:
    """
    Returns a list of IP addresses (as strings), given the value of an HTTP
    ``X-Forwarded-For`` (or ``WSGI HTTP_X_FORWARDED_FOR``) header.

    Args:
        value:
            the value of an HTTP ``X-Forwarded-For`` (or ``WSGI
            HTTP_X_FORWARDED_FOR``) header

    Returns:
        a list of IP address as strings

    See:
    - https://en.wikipedia.org/wiki/X-Forwarded-For
    - https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Forwarded-For  # noqa
    - NOT THIS: https://tools.ietf.org/html/rfc7239
    """
    if not value:
        return []
    return [x.strip() for x in value.split(",")]
    # ... separator is comma-space, but let's be liberal


def first_from_xff(value: str) -> str:
    """
    Returns the first IP address from an ``X-Forwarded-For`` header; see
    :func:`ip_addresses_from_xff`.

    Args:
        value:
            the value of an HTTP ``X-Forwarded-For`` (or ``WSGI
            HTTP_X_FORWARDED_FOR``) header

    Returns:
        an IP address as a string, or ``''`` if none is found

    """
    ip_addresses = ip_addresses_from_xff(value)
    if not ip_addresses:
        return ''
    return ip_addresses[0]  # leftmost


# =============================================================================
# Middleware to set SCRIPT_NAME environment variable etc. when behind a
# reverse proxy.
# =============================================================================

# noinspection HttpUrlsUsage
EXAMPLE_APACHE_REVERSE_PROXY_CONFIG = """

    # =========================================================================
    # Mount a WSGI application, using CamCOPS as an example
    # =========================================================================
    # This WSGI application is served by a SEPARATE web server (e.g. CherryPy);
    # Apache just needs to pass information to and fro, and handle the HTTPS
    # aspects.

        # ---------------------------------------------------------------------
        # 1. Proxy requests to the external server and back, and allow access.
        # ---------------------------------------------------------------------
        # ... either via port 8000
        # ... or, better, socket /tmp/.camcops.sock
        # NOTES
        # - Don't specify trailing slashes.
        #   If you do, http://host/camcops will fail, though
        #              http://host/camcops/ will succeed.
        # - Using a socket
        #   - this requires Apache 2.4.9, and passes after the '|' character a
        #     URL that determines the Host: value of the request; see
        #       https://httpd.apache.org/docs/trunk/mod/mod_proxy.html#proxypass
        #   - The Django debug toolbar will then require the bizarre entry in
        #     the Django settings: INTERNAL_IPS = ("b''", ) -- i.e. the string
        #     value of "b''", not an empty bytestring.
        # - Ensure that you put the CORRECT PROTOCOL (e.g. https) in the rules
        #   below.

        # (a) Proxy

        # ... via a port
        # Note the use of "http" (reflecting the backend), not https (like the
        # front end).
    ProxyPass /camcops http://127.0.0.1:8000 retry=0
    ProxyPassReverse /camcops http://127.0.0.1:8000 retry=0

        # ... or via a socket (Apache 2.4.9 and higher)
    # ProxyPass /camcops unix:/tmp/.camcops.sock|https://localhost retry=0
    # ProxyPassReverse /camcops unix:/tmp/.camcops.sock|https://localhost retry=0

        # (b) Allow proxy over SSL.
        # Without this, you will get errors like:
        #   ... SSL Proxy requested for wombat:443 but not enabled [Hint: SSLProxyEngine]
        #   ... failed to enable ssl support for 0.0.0.0:0 (httpd-UDS)
    SSLProxyEngine on

    <Location /camcops>
            # (c) Allow access
        Require all granted

            # (d) Tell the proxied application that we are using HTTPS:
            # ... https://stackoverflow.com/questions/16042647
            # Enable mod_headers (e.g. "sudo a2enmod headers") and:
        RequestHeader set X-Forwarded-Proto https
        RequestHeader set X-Script-Name /camcops
    </Location>

        # ---------------------------------------------------------------------
        # 2. Serve static files
        # ---------------------------------------------------------------------
        # a) offer them at the appropriate URL
        # b) provide permission

    Alias /camcops/static/ /usr/share/camcops/server/static/

    #   Change this: aim the alias at your own institutional logo.
    # Alias /camcops/static/logo_local.png /usr/share/camcops/server/static/logo_local.png

    <Directory /usr/share/camcops/server/static>
        Require all granted
    </Directory>

"""  # noqa


class ReverseProxiedConfig(object):
    """
    Class to hold information about a reverse proxy configuration.
    """
    def __init__(self,
                 trusted_proxy_headers: List[str] = None,
                 http_host: str = None,
                 remote_addr: str = None,
                 script_name: str = None,
                 server_name: str = None,
                 server_port: int = None,
                 url_scheme: str = None,
                 rewrite_path_info: bool = False) -> None:
        # noinspection HttpUrlsUsage
        """
        Args:
            trusted_proxy_headers:
                list of headers, from
                :const:`ReverseProxiedMiddleware.ALL_CANDIDATES`, that the
                middleware will treat as trusted and obey. All others from this
                list will be stripped.

            http_host:
                Value to write to the ``HTTP_HOST`` WSGI variable. If not
                specified, an appropriate trusted header will be used (if there
                is one).

            remote_addr:
                ... similarly for ``REMOTE_ADDR``

            script_name:
                ... similarly for ``SCRIPT_NAME``

            server_name:
                ... similarly for ``SERVER_NAME``

            server_port:
                ... similarly for ``SERVER_PORT``

            url_scheme:
                ... similarly for ``URL_SCHEME`` (e.g. ``"https"``)

            rewrite_path_info:
                If ``True``, then if the middleware sets ``SCRIPT_NAME`` and
                ``PATH_INFO`` starts with ``SCRIPT_NAME``, the ``SCRIPT_NAME``
                will be stripped off the front of ``PATH_INFO``.

                This is appropriate for front-end web servers that fail to
                rewrite the incoming URL properly. (Do not use for Apache with
                ``ProxyPass``; ``ProxyPass`` rewrites the URLs properly for
                you.)

                ... as per e.g. http://flask.pocoo.org/snippets/35/
        """
        self.trusted_proxy_headers = []  # type: List[str]
        if trusted_proxy_headers:
            for x in trusted_proxy_headers:
                h = x.upper()
                if h in ReverseProxiedMiddleware.ALL_CANDIDATES:
                    self.trusted_proxy_headers.append(h)
        self.http_host = http_host
        self.remote_addr = remote_addr
        self.script_name = script_name.rstrip("/") if script_name else ""
        self.server_name = server_name
        self.server_port = str(server_port) if server_port is not None else ""
        self.url_scheme = url_scheme.lower() if url_scheme else ""
        self.rewrite_path_info = rewrite_path_info

    def necessary(self) -> bool:
        """
        Is any special handling (e.g. the addition of
        :class:`ReverseProxiedMiddleware`) necessary for thie config?
        """
        return any([
            self.trusted_proxy_headers,
            self.http_host,
            self.remote_addr,
            self.script_name,
            self.server_name,
            self.server_port,
            self.url_scheme,
            self.rewrite_path_info,
        ])


class ReverseProxiedMiddleware(object):
    # noinspection HttpUrlsUsage
    """
    WSGI middleware to set the ``SCRIPT_NAME`` and ``PATH_INFO`` WSGI
    environment variables (etc.) correctly when behind a reverse proxy.

    Note that the WSGI environment variables ``HTTP_*`` are clones of HTTP
    headers; for example, ``X-Forwarded-For`` in HTTP becomes
    ``HTTP_X_FORWARDED_FOR`` in WSGI.

    See also:

    - http://flask.pocoo.org/snippets/35/
    - http://blog.macuyiko.com/post/2016/fixing-flask-url_for-when-behind-mod_proxy.html
    - http://alex.eftimie.ro/2013/03/21/how-to-run-flask-application-in-a-subpath-using-apache-mod_proxy/
    - http://modwsgi.readthedocs.io/en/develop/release-notes/version-4.4.9.html
    - https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers

    """  # noqa

    CANDIDATES_HTTP_HOST = [
        # These are variables that may contain a value for HTTP_HOST.
        WsgiEnvVar.HTTP_X_HOST,
        WsgiEnvVar.HTTP_X_FORWARDED_HOST
    ]
    CANDIDATES_SERVER_PORT = [
        # These are variables that may contain a value for SERVER_PORT.
        WsgiEnvVar.HTTP_X_FORWARDED_PORT
    ]
    CANDIDATES_REMOTE_ADDR = [
        # These are variables that may contain a value for REMOTE_ADDR.
        # However, they differ:
        WsgiEnvVar.HTTP_X_FORWARDED_FOR,  # may contain many values; first is taken  # noqa
        WsgiEnvVar.HTTP_X_REAL_IP  # may contain only one
    ]
    _CANDIDATES_URL_SCHEME_GIVING_PROTOCOL = [
        # These are variables whose values might be "http" or "https",
        # relevant to wsgi.url_scheme.
        WsgiEnvVar.HTTP_X_FORWARDED_PROTO,
        WsgiEnvVar.HTTP_X_FORWARDED_PROTOCOL,
        WsgiEnvVar.HTTP_X_FORWARDED_SCHEME,
        WsgiEnvVar.HTTP_X_SCHEME
    ]
    _CANDIDATES_URL_SCHEME_INDICATING_HTTPS = [
        # These are variables whose values might be "On", "True", or "1",
        # to indicate the use of HTTPS/SSL, relevant to wsgi.url_scheme.
        WsgiEnvVar.HTTP_X_FORWARDED_HTTPS,
        WsgiEnvVar.HTTP_X_FORWARDED_SSL,
        WsgiEnvVar.HTTP_X_HTTPS,
    ]
    CANDIDATES_URL_SCHEME = (
        # All variables that may inform wsgi.url_scheme
        _CANDIDATES_URL_SCHEME_GIVING_PROTOCOL +
        _CANDIDATES_URL_SCHEME_INDICATING_HTTPS
    )
    CANDIDATES_SCRIPT_NAME = [
        # These are variables that may contain a value for SCRIPT_NAME.
        WsgiEnvVar.HTTP_X_SCRIPT_NAME,
        WsgiEnvVar.HTTP_X_FORWARDED_SCRIPT_NAME
    ]
    CANDIDATES_SERVER_NAME = [
        # These are variables that may contain a value for SERVER_NAME.
        WsgiEnvVar.HTTP_X_FORWARDED_SERVER
    ]
    ALL_CANDIDATES = (
        # All variables of interest.
        CANDIDATES_HTTP_HOST +
        CANDIDATES_SERVER_PORT +
        CANDIDATES_REMOTE_ADDR +
        _CANDIDATES_URL_SCHEME_GIVING_PROTOCOL +
        _CANDIDATES_URL_SCHEME_INDICATING_HTTPS +
        CANDIDATES_SCRIPT_NAME +
        CANDIDATES_SERVER_NAME
    )
    SCHEME_HTTPS = 'https'
    TRUE_VALUES_LOWER_CASE = ["on", "true", "1"]

    def __init__(self,
                 app: TYPE_WSGI_APP,
                 config: ReverseProxiedConfig,
                 debug: bool = False) -> None:
        self.app = app
        self.config = config
        self.debug = debug

        self.vars_host = [x for x in self.CANDIDATES_HTTP_HOST
                          if x in config.trusted_proxy_headers]
        self.vars_addr = [x for x in self.CANDIDATES_REMOTE_ADDR
                          if x in config.trusted_proxy_headers]
        self.vars_script = [x for x in self.CANDIDATES_SCRIPT_NAME
                            if x in config.trusted_proxy_headers]
        self.vars_server = [x for x in self.CANDIDATES_SERVER_NAME
                            if x in config.trusted_proxy_headers]
        self.vars_port = [x for x in self.CANDIDATES_SERVER_PORT
                          if x in config.trusted_proxy_headers]
        self.vars_scheme_a = [
            x for x in self._CANDIDATES_URL_SCHEME_GIVING_PROTOCOL
            if x in config.trusted_proxy_headers
        ]
        self.vars_scheme_b = [
            x for x in self._CANDIDATES_URL_SCHEME_INDICATING_HTTPS
            if x in config.trusted_proxy_headers
        ]

        if self.debug:
            log.debug("ReverseProxiedMiddleware installed")
            self._report(WsgiEnvVar.HTTP_HOST, config.http_host,
                         self.vars_host)
            self._report(WsgiEnvVar.REMOTE_ADDR, config.remote_addr,
                         self.vars_addr)
            self._report(WsgiEnvVar.SCRIPT_NAME, config.script_name,
                         self.vars_script)
            if config.script_name or self.vars_script:
                log.debug("... which will also affect WSGI environment "
                          "variable {}", WsgiEnvVar.PATH_INFO)
            self._report(WsgiEnvVar.SERVER_NAME, config.server_name,
                         self.vars_server)
            self._report(WsgiEnvVar.SERVER_PORT, config.server_port,
                         self.vars_port)
            self._report(WsgiEnvVar.WSGI_URL_SCHEME, config.url_scheme,
                         self.vars_scheme_a + self.vars_scheme_b)

    @staticmethod
    def _report(option: str, value: str, envvars: List[str]) -> None:
        if value:
            log.debug("... WSGI environment variable {} will be set to {}",
                      option, value)
        elif envvars:
            log.debug("... WSGI environment variable {} will be set according "
                      "to reflect environment variables {!r} in "
                      "incoming requests",
                      option, envvars)
        else:
            log.debug("... WSGI environment variable {} will not be changed",
                      option)

    @classmethod
    def _get_first(cls,
                   environ: TYPE_WSGI_ENVIRON,
                   envvars: List[str],
                   keys_to_keep: List[str],
                   as_remote_addr: bool = False) -> str:
        for k in envvars:
            value = environ.get(k, '')
            if value:
                keys_to_keep.append(k)
                # Oddity for REMOTE_ADDR and X-Forwarded-For:
                if as_remote_addr and k == WsgiEnvVar.HTTP_X_FORWARDED_FOR:
                    value = first_from_xff(value)
                return value
        return ''

    @classmethod
    def _proto_if_one_true(cls,
                           environ: TYPE_WSGI_ENVIRON,
                           envvars: List[str],
                           keys_to_keep: List[str]) -> str:
        for k in envvars:
            value = environ.get(k, '')
            if value.lower() in cls.TRUE_VALUES_LOWER_CASE:
                keys_to_keep.append(k)
                return cls.SCHEME_HTTPS
        return ''

    def __call__(self,
                 environ: TYPE_WSGI_ENVIRON,
                 start_response: TYPE_WSGI_START_RESPONSE) \
            -> TYPE_WSGI_APP_RESULT:
        # noinspection HttpUrlsUsage
        """
        -----------------------------------------------------------------------
        REWRITING THE HOST (setting HTTP_HOST):
        -----------------------------------------------------------------------

        If you don't rewrite the host, the Pyramid debug toolbar will get
        things a bit wrong. An example:
            http://127.0.0.1:80/camcops
        is proxied by Apache to
            http://127.0.0.7:8000/camcops

        In that situation, HTTP_HOST will be '127.0.0.1:8000', and so the
        Pyramid debug toolbar will start asking the web browser to go to
            http://127.0.0.1:8000/camcops/_debug_toolbar/...
        ... which is wrong (it's a reference to the "internal" site).

        If you allow the host to be rewritten, then you get a sensible
        reference e.g. to
            http://wombat/camcops/_debug_toolbar/...

        Should we be looking at HTTP_X_FORWARDED_HOST or
        HTTP_X_FORWARDED_SERVER?
        See https://github.com/omnigroup/Apache/blob/master/httpd/modules/proxy/mod_proxy_http.c  # noqa
        ... and let's follow mod_wsgi.

        -----------------------------------------------------------------------
        HTTP_HOST versus SERVER_NAME
        -----------------------------------------------------------------------
        https://stackoverflow.com/questions/2297403/what-is-the-difference-between-http-host-and-server-name-in-php  # noqa

        -----------------------------------------------------------------------
        REWRITING THE PROTOCOL
        -----------------------------------------------------------------------
        Consider how we get here. For example, we may have this sequence:

        .. code-block:: none

            user's web browser
            -> Apache front-end web server via HTTPS on port 443
                -> ProxyPass/ProxyPassReverse
            -> CherryPy server via HTTP on port 8000 or via a Unix socket
                -> ...
                -> cherrypy/wsgiserver/__init__.py,
                    WSGIGateway_10.get_environ()
                    ... which creates a WSGI environment from an HTTP request.

        So if you want to see what's coming by way of raw headers, put this
        in at the end of that get_environ() function:

        .. code-block:: python

            from pprint import pformat; import logging; log = logging.getLogger(__name__); log.critical("Request headers:\n" + pformat(req.inheaders))

        """  # noqa
        if self.debug:
            log.debug("Starting WSGI environment: \n{}", pformat(environ))
            oldenv = environ.copy()
        keys_to_keep = []  # type: List[str]
        config = self.config

        # ---------------------------------------------------------------------

        # HTTP_HOST
        http_host = (
            config.http_host or  # manually specified: top priority. Otherwise:
            self._get_first(environ, self.vars_host, keys_to_keep)
        )
        if http_host:
            environ[WsgiEnvVar.HTTP_HOST] = http_host

        # REMOTE_ADDR
        remote_addr = (
            config.remote_addr or
            self._get_first(environ, self.vars_addr, keys_to_keep,
                            as_remote_addr=True)
        )
        if remote_addr:
            environ[WsgiEnvVar.REMOTE_ADDR] = remote_addr

        # SCRIPT_NAME, PATH_INFO
        script_name = (
            config.script_name or
            self._get_first(environ, self.vars_script, keys_to_keep)
        )
        if script_name:
            environ[WsgiEnvVar.SCRIPT_NAME] = script_name
            path_info = environ[WsgiEnvVar.PATH_INFO]
            if config.rewrite_path_info and path_info.startswith(script_name):
                newpath = path_info[len(script_name):]
                if not newpath:  # e.g. trailing slash omitted from incoming path  # noqa
                    newpath = "/"
                environ[WsgiEnvVar.PATH_INFO] = newpath

        # SERVER_NAME
        server_name = (
            config.server_name or
            self._get_first(environ, self.vars_server, keys_to_keep)
        )
        if server_name:
            environ[WsgiEnvVar.SERVER_NAME] = server_name

        # SERVER_PORT
        server_port = (
            config.server_port or
            self._get_first(environ, self.vars_port, keys_to_keep)
        )
        if server_port:
            environ[WsgiEnvVar.SERVER_PORT] = server_port

        # wsgi.url_scheme
        url_scheme = (
            config.url_scheme or  # manually specified: top priority. Otherwise:
            self._get_first(environ, self.vars_scheme_a, keys_to_keep) or
            self._proto_if_one_true(environ, self.vars_scheme_b, keys_to_keep)
        )
        if url_scheme:
            url_scheme = url_scheme.lower()
            environ[WsgiEnvVar.WSGI_URL_SCHEME] = url_scheme

        # ---------------------------------------------------------------------

        # As per mod_wsgi, we delete unused and untrusted keys.
        delete_keys(environ,
                    keys_to_delete=self.ALL_CANDIDATES,
                    keys_to_keep=keys_to_keep)
        if self.debug:
            # noinspection PyUnboundLocalVariable
            changes = dict_diff(oldenv, environ)
            log.debug("Changes to WSGI environment: \n{}", pformat(changes))
        return self.app(environ, start_response)



#!/usr/bin/env python
# cardinal_pythonlib/wsgi/constants.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Miscellany to help with WSGI work.**

"""

from types import TracebackType
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Type

# =============================================================================
# Type hints for WSGI
# =============================================================================

# https://www.python.org/dev/peps/pep-0333/
TYPE_WSGI_ENVIRON = Dict[str, str]
TYPE_WSGI_STATUS = str
TYPE_WSGI_RESPONSE_HEADERS = List[Tuple[str, str]]
TYPE_WSGI_START_RESP_RESULT = Callable[[str], None]  # call with e.g. write(body_data)  # noqa
TYPE_WSGI_EXC_INFO = Tuple[
    # https://docs.python.org/3/library/sys.html#sys.exc_info
    Optional[Type[BaseException]],  # type
    Optional[BaseException],  # value
    Optional[TracebackType]  # traceback
]
TYPE_WSGI_START_RESPONSE = Callable[
    # There is an optional third parameter to start_response():
    [TYPE_WSGI_STATUS,  # status
     TYPE_WSGI_RESPONSE_HEADERS,  # headers
     Optional[TYPE_WSGI_EXC_INFO]],  # exc_info
    TYPE_WSGI_START_RESP_RESULT
]
TYPE_WSGI_APP_RESULT = Iterable[bytes]
# ... must be BYTE STRINGS, not str; see
# https://www.python.org/dev/peps/pep-0333/#unicode-issues
# and also
# https://github.com/fgallaire/wsgiserver/issues/2
TYPE_WSGI_APP = Callable[[TYPE_WSGI_ENVIRON, TYPE_WSGI_START_RESPONSE],
                         TYPE_WSGI_APP_RESULT]


# =============================================================================
# Constants
# =============================================================================

class WsgiEnvVar(object):
    """
    WSGI environment variable names.

    For core ones, see https://wsgi.readthedocs.io/en/latest/definitions.html
    """
    CONTENT_LENGTH = "CONTENT_LENGTH"  # [1]
    CONTENT_TYPE = "CONTENT_TYPE"  # [1]
    HTTP_HOST = "HTTP_HOST"  # [2]
    HTTP_X_FORWARDED_FOR = "HTTP_X_FORWARDED_FOR"  # [2]
    HTTP_X_REAL_IP = "HTTP_X_REAL_IP"  # [7]
    HTTP_X_FORWARDED_HTTPS = "HTTP_X_FORWARDED_HTTPS"  # [7]
    HTTP_X_FORWARDED_HOST = "HTTP_X_FORWARDED_HOST"  # [4]
    HTTP_X_FORWARDED_SCRIPT_NAME = "HTTP_X_FORWARDED_SCRIPT_NAME"  # [7]
    HTTP_X_FORWARDED_SERVER = "HTTP_X_FORWARDED_SERVER"  # [4]
    HTTP_X_FORWARDED_PORT = "HTTP_X_FORWARDED_PORT"  # [7]
    HTTP_X_FORWARDED_PROTO = "HTTP_X_FORWARDED_PROTO"  # [6]
    HTTP_X_FORWARDED_PROTOCOL = "HTTP_X_FORWARDED_PROTOCOL"  # [6]
    HTTP_X_FORWARDED_SCHEME = "HTTP_X_FORWARDED_SCHEME"  # [7]
    HTTP_X_FORWARDED_SSL = "HTTP_X_FORWARDED_SSL"  # [7]
    HTTP_X_HOST = "HTTP_X_HOST"  # [7]
    HTTP_X_HTTPS = "HTTP_X_HTTPS"  # [7]
    HTTP_X_SCHEME = "HTTP_X_SCHEME"  # [5]
    HTTP_X_SCRIPT_NAME = "HTTP_X_SCRIPT_NAME"  # [5]
    PATH_INFO = "PATH_INFO"  # [1]
    QUERY_STRING = "QUERY_STRING"  # [1]
    REMOTE_ADDR = "REMOTE_ADDR"  # [1]
    REQUEST_METHOD = "REQUEST_METHOD"  # [1]
    SCRIPT_NAME = "SCRIPT_NAME"  # [1]
    SERVER_NAME = "SERVER_NAME"  # [1]
    SERVER_PORT = "SERVER_PORT"  # [1]
    SERVER_PROTOCOL = "SERVER_PROTOCOL"  # [1]
    WSGI_ERRORS = "wsgi.errors"  # [3]
    WSGI_INPUT = "wsgi.input"  # [3]
    WSGI_MULTIPROCESS = "wsgi.multiprocess"  # [3]
    WSGI_MULTITHREAD = "wsgi.multithread"  # [3]
    WSGI_RUN_ONCE = "wsgi.run_once"  # [3]
    WSGI_URL_SCHEME = "wsgi.url_scheme"  # [3]
    WSGI_VERSION = "wsgi.version"  # [3]

    # [1] Standard WSGI and standard CGI; must always be present;
    #     https://wsgi.readthedocs.io/en/latest/definitions.html
    #     https://www.python.org/dev/peps/pep-0333/
    #     https://en.wikipedia.org/wiki/Common_Gateway_Interface
    # [2] Standard WSGI as copies of standard HTTP request fields (thus,
    #     optional); https://wsgi.readthedocs.io/en/latest/definitions.html
    # [3] Also standard WSGI, but not CGI; must always be present.
    # [4] From non-standard but common HTTP request fields;
    #     https://en.wikipedia.org/wiki/List_of_HTTP_header_fields#Common_non-standard_request_fields  # noqa
    #     https://github.com/omnigroup/Apache/blob/master/httpd/modules/proxy/mod_proxy_http.c  # noqa
    # [5] Non-standard; Nginx-specific? Nonetheless, all "HTTP_" variables in
    #     WSGI should follow the HTTP request headers.
    # [6] Protocols (i.e. http versus https):
    #     https://stackoverflow.com/questions/16042647/whats-the-de-facto-standard-for-a-reverse-proxy-to-tell-the-backend-ssl-is-used  # noqa
    # [7] https://modwsgi.readthedocs.io/en/develop/release-notes/version-4.4.9.html  # noqa



#!/usr/bin/env python
# cardinal_pythonlib/argparse_func.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to help with argparse.**

"""

# noinspection PyProtectedMember
from argparse import (
    _HelpAction,
    _SubParsersAction,
    ArgumentDefaultsHelpFormatter,
    ArgumentParser,
    ArgumentTypeError,
    Namespace,
    RawDescriptionHelpFormatter,
)
from typing import Any, Dict, List, Type


# =============================================================================
# Argparse actions
# =============================================================================

class ShowAllSubparserHelpAction(_HelpAction):
    """
    Class to serve as the ``action`` for an ``argparse`` top-level parser that
    shows help for all subparsers. As per

    https://stackoverflow.com/questions/20094215/argparse-subparser-monolithic-help-output
    """  # noqa

    def __call__(self,
                 parser: ArgumentParser,
                 namespace: Namespace,
                 values: List[Any],  # ?
                 option_string: str = None) -> None:
        # 1. Print top-level help
        parser.print_help()
        sep = "=" * 79  # "-" less helpful when using grep for "--option"!

        # 2. Print help for all subparsers
        # noinspection PyProtectedMember
        subparsers_actions = [
            action for action in parser._actions
            if isinstance(action, _SubParsersAction)
        ]  # type: List[_SubParsersAction]
        messages = [""]  # type: List[str]
        for subparsers_action in subparsers_actions:
            for choice, subparser in subparsers_action.choices.items():
                messages.append(sep)
                messages.append(f"Help for command '{choice}'")
                messages.append(sep)
                messages.append(subparser.format_help())
        print("\n".join(messages))

        parser.exit()


# =============================================================================
# Argparse formatters
# =============================================================================

class RawDescriptionArgumentDefaultsHelpFormatter(
        ArgumentDefaultsHelpFormatter,
        RawDescriptionHelpFormatter):
    """
    Combines the features of

    - :class:`RawDescriptionHelpFormatter` -- don't mangle the description
    - :class:`ArgumentDefaultsHelpFormatter` -- print argument defaults
    """
    pass


# =============================================================================
# Argparse types/checkers
# =============================================================================

def str2bool(v: str) -> bool:
    """
    ``argparse`` type that maps strings in case-insensitive fashion like this:

    .. code-block:: none

        argument strings                value
        ------------------------------- -----
        'yes', 'true', 't', 'y', '1'    True
        'no', 'false', 'f', 'n', '0'    False

    From
    https://stackoverflow.com/questions/15008758/parsing-boolean-values-with-argparse

    Specimen usage:

    .. code-block:: python

        parser.add_argument(
            "--nice", type=str2bool, nargs='?',
            const=True,  # if --nice is present with no parameter
            default=NICE,  # if the argument is entirely absent
            help="Activate nice mode.")

    """  # noqa
    lv = v.lower()
    if lv in ('yes', 'true', 't', 'y', '1'):
        return True
    elif lv in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise ArgumentTypeError('Boolean value expected.')


def positive_int(value: str) -> int:
    """
    ``argparse`` argument type that checks that its value is a positive
    integer.
    """
    try:
        ivalue = int(value)
        assert ivalue > 0
    except (AssertionError, TypeError, ValueError):
        raise ArgumentTypeError(
            f"{value!r} is an invalid positive int")
    return ivalue


def nonnegative_int(value: str) -> int:
    """
    ``argparse`` argument type that checks that its value is a non-negative
    integer.
    """
    try:
        ivalue = int(value)
        assert ivalue >= 0
    except (AssertionError, TypeError, ValueError):
        raise ArgumentTypeError(
            f"{value!r} is an invalid non-negative int")
    return ivalue


def percentage(value: str) -> float:
    """
    ``argparse`` argument type that checks that its value is a percentage (in
    the sense of a float in the range [0, 100]).
    """
    try:
        fvalue = float(value)
        assert 0 <= fvalue <= 100
    except (AssertionError, TypeError, ValueError):
        raise ArgumentTypeError(
            f"{value!r} is an invalid percentage value")
    return fvalue


class MapType(object):
    """
    ``argparse`` type maker that maps strings to a dictionary (map).
    """

    def __init__(self,
                 map_separator: str = ":",
                 pair_separator: str = ",",
                 strip: bool = True,
                 from_type: Type = str,
                 to_type: Type = str) -> None:
        """
        Args:
            map_separator:
                string that separates the "from" and "to" members of a pair
            pair_separator:
                string that separates different pairs
            strip:
                strip whitespace after splitting?
            from_type:
                type to coerce "from" values to; e.g. ``str``, ``int``
            to_type:
                type to coerce "to" values to; e.g. ``str``, ``int``
        """
        self.map_separator = map_separator
        self.pair_separator = pair_separator
        self.strip = strip
        self.from_type = from_type
        self.to_type = to_type

    def __call__(self, value: str) -> Dict:
        result = {}
        pairs = value.split(self.pair_separator)
        for pair in pairs:
            from_str, to_str = pair.split(self.map_separator)
            if self.strip:
                from_str = from_str.strip()
                to_str = to_str.strip()
            try:
                from_val = self.from_type(from_str)
            except (TypeError, ValueError):
                raise ArgumentTypeError(
                    f"{from_str!r} cannot be converted to type "
                    f"{self.from_type!r}")
            try:
                to_val = self.to_type(to_str)
            except (TypeError, ValueError):
                raise ArgumentTypeError(
                    f"{to_str!r} cannot be converted to type "
                    f"{self.to_type!r}")
            result[from_val] = to_val
        return result



#!/usr/bin/env python
# cardinal_pythonlib/tools/convert_mdb_to_mysql.py

# noinspection HttpUrlsUsage
"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

***Converts a .MDB file (Microsoft Access) database to MySQL, copying structure
and data.***

Uses the ``mdb-tools`` package.

Under Ubuntu, the following packages are required:

.. code-block: bash

    sudo apt-get install mdbtools mysql-server mysql-client mysql-admin
    
though you may also want these, if you're planning to use Python/ODBC with
MySQL:

.. code-block: bash

    sudo apt-get install mysql-navigator libmdbtools libmdbodbc unixodbc python-mysqldb python-pyodbc

What works:

- schema copied
- data copied

What doesn't work:

- indexes are not described by mdb-schema, so these must be recreated
- relationships are not supported by mdb-schema, so these must be recreated

We'll do this with calls to other command-line tools. See
http://nialldonegan.me/2007/03/10/converting-microsoft-access-mdb-into-csv-or-mysql-in-linux/

History

- REVISED 1 Jan 2013: mdb-schema syntax has changed (-S option gone).
  See https://github.com/brianb/mdbtools.
- REVISED 16 Jan 2017: conversion to Python 3.
- Fixed a bit more, 2020-01-19. Also type hinting.

"""  # noqa

import argparse
import getpass
import logging
import re
import subprocess
import sys
import tempfile
from typing import Any, Callable, List, Optional, Union

from cardinal_pythonlib.logs import main_only_quicksetup_rootlogger

log = logging.getLogger(__name__)


def get_command_output(
        arglist: List[str],
        encoding: str = sys.getdefaultencoding()) -> Optional[str]:
    log.debug("Executing command: {}".format(arglist))
    result = subprocess.check_output(arglist)
    return result.decode(encoding) if result else None


def get_pipe_series_output(
        commands: List[List[str]],
        inputstr: str = None,
        encoding: str = sys.getdefaultencoding()) -> Optional[str]:
    processes = []
    log.debug("Executing commands: {}".format(
        " -> ".join(" ".join(arglist) for arglist in commands))
    )
    for i in range(len(commands)):
        arglist = commands[i]
        if i == 0:  # first processes
            processes.append(subprocess.Popen(arglist,
                                              stdin=subprocess.PIPE,
                                              stdout=subprocess.PIPE))
        else:  # subsequent ones
            processes.append(subprocess.Popen(arglist,
                                              stdin=processes[i - 1].stdout,
                                              stdout=subprocess.PIPE))
    stdout_bytes = processes[-1].communicate(inputstr.encode(encoding))[0]
    # communicate() returns a tuple; 0=stdout, 1=stderr; so this returns stdout
    return stdout_bytes.decode(encoding) if stdout_bytes else None


def replace_type_in_sql(sql: str, fromstr: str, tostr: str) -> str:
    whitespaceregroup = r"([\ \t\n]+)"
    whitespaceorcommaregroup = r"([\ \t\),\n]+)"
    rg1 = r"\g<1>"
    rg2 = r"\g<2>"
    return re.sub(whitespaceregroup + fromstr + whitespaceorcommaregroup,
                  rg1 + tostr + rg2, sql,
                  0,
                  re.MULTILINE | re.IGNORECASE)


class PasswordPromptAction(argparse.Action):
    """
    This provides a command-line option to provide a password or read it
    interactively.
    
    Use it like this:
    
    .. code-block:: python
    
        parser.add_argument(
            '--password', type=str, action=PasswordPromptAction,
            help="MySQL password")
     
    Modified from
    https://stackoverflow.com/questions/27921629/python-using-getpass-with-argparse 
    """  # noqa
    # noinspection PyShadowingBuiltins
    def __init__(self,
                 option_strings: List[str],
                 dest: str = None,
                 nargs: Union[int, str] = "?",  # 0 or 1
                 default: Any = None,
                 required: bool = False,
                 type: Callable[[str], Any] = None,
                 metavar: str = None,
                 help: str = None) -> None:
        super(PasswordPromptAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=nargs,
            default=default,
            required=required,
            metavar=metavar,
            type=type,
            help=help)

    def __call__(self,
                 parser: argparse.ArgumentParser,
                 args: argparse.Namespace,
                 values: Union[None, str, List[str]],
                 option_string: str = None) -> None:
        if isinstance(values, list) and len(values) == 1:
            password = values[0]
        else:
            password = values
        if not password:
            password = getpass.getpass()
        setattr(args, self.dest, password)


def main() -> None:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        'mdbfile', type=str,
        help="Microsoft Access .MDB file to read")
    parser.add_argument(
        '--schemaonly', action='store_true',
        help="Print schema SQL only (otherwise: will create MySQL database and"
             " write schema and data)")
    parser.add_argument(
        '--host', type=str, default='127.0.0.1',  # not "localhost"
        help="MySQL hostname or IP address")
    parser.add_argument(
        '--port', type=int, default=3306,
        help="MySQL port number")
    parser.add_argument(
        '--user', type=str, default='root',
        help="MySQL username")
    parser.add_argument(
        '--password', type=str, action=PasswordPromptAction,
        help="MySQL password")
    parser.add_argument(
        '--mysqldb', type=str,
        help="MySQL database to use")
    parser.add_argument(
        '--create', action="store_true",
        help="Create database?")
    args = parser.parse_args()

    if not args.schemaonly and not args.mysqldb:
        raise ValueError("Must specify --mysqldb, unless --schemaonly is used")

    main_only_quicksetup_rootlogger(level=logging.DEBUG)

    # -------------------------------------------------------------------------
    log.info("Getting list of tables")
    # -------------------------------------------------------------------------
    tablecmd = ["mdb-tables", "-1", args.mdbfile]
    # ... -1: one per line (or table names with spaces will cause confusion)
    tables = get_command_output(tablecmd).splitlines()
    tables.sort(key=lambda s: s.lower())
    log.info("Tables: {}".format(tables))

    # -------------------------------------------------------------------------
    log.info("Fetching schema definition")
    # -------------------------------------------------------------------------
    schemacmd = [
        "mdb-schema",  # from Ubuntu package mdbtools
        args.mdbfile,  # database
        "mysql",  # backend
        "--no-drop-table",  # don't issue DROP TABLE statements (default)
        "--not-null",  # issue NOT NULL constraints (default)
        "--no-default-values",  # don't issue DEFAULT values (default)
        "--indexes",  # export INDEX statements (default)
        "--relations",  # export foreign key constraints (default)
    ]
    schemasyntax = get_command_output(schemacmd)

    # -------------------------------------------------------------------------
    log.info("Converting any schema syntax oddities to MySQL syntax")
    # -------------------------------------------------------------------------
    # JAN 2013: Since my previous script, mdb-schema's mysql dialect has got
    # much better. So, not much to do here.

    # "COMMENT ON COLUMN" produced by mdb-schema and rejected by MySQL:
    schemasyntax = re.sub("^COMMENT ON COLUMN.*$", "", schemasyntax, 0,
                          re.MULTILINE)

    log.info("Schema syntax:")
    print(schemasyntax)

    # -------------------------------------------------------------------------
    # Done?
    # -------------------------------------------------------------------------
    if args.schemaonly:
        return

    # -------------------------------------------------------------------------
    log.info("Creating new database")
    # -------------------------------------------------------------------------
    createmysqldbcmd = [
        "mysqladmin",
        "create", args.mysqldb,
        "--host={}".format(args.host),
        "--port={}".format(args.port),
        "--user={}".format(args.user),
        "--password={}".format(args.password)
    ]
    # We could omit the actual password and the user would be prompted, but we
    # need to send it this way later (see below), so this is not a huge
    # additional security weakness!
    # Linux/MySQL helpfully obscures the password in the "ps" list.
    log.info(get_command_output(createmysqldbcmd))

    # -------------------------------------------------------------------------
    log.info("Sending schema command to MySQL")
    # -------------------------------------------------------------------------
    mysqlcmd = [
        "mysql",
        "--host={}".format(args.host),
        "--port={}".format(args.port),
        "--database={}".format(args.mysqldb),
        "--user={}".format(args.user),
        "--password={}".format(args.password)
    ]
    # Regrettably, we need the password here, as stdin will come from a pipe
    print(get_pipe_series_output([mysqlcmd], schemasyntax))

    # -------------------------------------------------------------------------
    log.info("Copying data to MySQL")
    # -------------------------------------------------------------------------
    # For the data, we won't store the intermediate stuff in Python's memory,
    # 'cos it's vast; I had one odd single-character mutation
    # from "TimeInSession_ms" to "TimeInSession_mc" at row 326444 (perhaps
    # therefore 37Mb or so into a long string).
    # And I was trying to export ~1m records in that table alone.
    # We'll use pipes instead and let the OS deal with the memory management.

    # ... BUT (Jan 2013): now mdb-tools is better, text-processing not
    # necessary - can use temporary disk file
    # Turns out the bottleneck is the import to MySQL, not the export from MDB.
    # So see http://dev.mysql.com/doc/refman/5.5/en/optimizing-innodb-bulk-data-loading.html  # noqa
    # The massive improvement is by disabling autocommit. (Example source
    # database is 208M; largest table here is 554M as a textfile; it has
    # 1,686,075 rows.) This improvement was from 20 Hz to the whole database
    # in a couple of minutes (~13 kHz). Subsequent export from MySQL: takes a
    # second or two to write whole DB (177M textfile).

    for t in tables:
        log.info("Processing table {}".format(t))
        exportcmd = [
            'mdb-export',
            '-I', 'mysql',  # -I backend: INSERT statements, not CSV

            # MySQL's DATETIME field has this format: "YYYY-MM-DD HH:mm:SS"
            # so we want this from the export:
            '-D', '%Y-%m-%d %H:%M:%S',  # -D: date format
            # ... don't put any extra quotes around it.

            args.mdbfile,  # database
            t  # table
        ]
        with tempfile.NamedTemporaryFile(
                mode="wt",
                encoding=sys.getdefaultencoding()) as outfile:
            print("SET autocommit=0;", file=outfile)
            subprocess.call(exportcmd, stdout=outfile)
            print("\nCOMMIT;", file=outfile)
            outfile.flush()
            outfile.seek(0)
            subprocess.call(mysqlcmd, stdin=outfile)

    log.info("Finished.")


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# cardinal_pythonlib/profile.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Profiling assistance functions.**

"""

import cProfile
from typing import Any, Callable


def do_cprofile(func: Callable, sort: str = "tottime") -> Callable:
    """
    Print profile stats to screen. To be used as a decorator for the function
    or method you want to profile. For example:

    .. code-block:: python

        profiled_func = do_cprofile(original_func)
        profiled_func(args_to_original_func)

    """
    def profiled_func(*args, **kwargs) -> Any:
        profile = cProfile.Profile()
        try:
            profile.enable()
            result = func(*args, **kwargs)
            profile.disable()
            return result
        finally:
            profile.print_stats(sort=sort)
    return profiled_func



#!/usr/bin/env python
# cardinal_pythonlib/sort.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Support functions for sorting.**

"""

from functools import partial, total_ordering
import re
from typing import Any, List, Union


# =============================================================================
# Natural sorting, e.g. for COM ports
# =============================================================================
# https://stackoverflow.com/questions/5967500/how-to-correctly-sort-a-string-with-a-number-inside  # noqa

def atoi(text: str) -> Union[int, str]:
    """
    Converts strings to integers if they're composed of digits; otherwise
    returns the strings unchanged. One way of sorting strings with numbers;
    it will mean that ``"11"`` is more than ``"2"``.
    """
    return int(text) if text.isdigit() else text


def natural_keys(text: str) -> List[Union[int, str]]:
    """
    Sort key function.
    Returns text split into string/number parts, for natural sorting; as per
    https://stackoverflow.com/questions/5967500/how-to-correctly-sort-a-string-with-a-number-inside

    Example (as per the source above):

    .. code-block:: python

        >>> from cardinal_pythonlib.sort import natural_keys
        >>> alist=[
        ...     "something1",
        ...     "something12",
        ...     "something17",
        ...     "something2",
        ...     "something25",
        ...     "something29"
        ... ]
        >>> alist.sort(key=natural_keys)
        >>> alist
        ['something1', 'something2', 'something12', 'something17', 'something25', 'something29']

    """  # noqa
    return [atoi(c) for c in re.split(r'(\d+)', text)]


# =============================================================================
# Sorting where None counts as the minimum
# =============================================================================

@total_ordering
class MinType(object):
    """
    An object that compares less than anything else.
    """
    def __le__(self, other: Any) -> bool:
        return True

    def __eq__(self, other: Any) -> bool:
        return self is other

    def __str__(self) -> str:
        return "MinType"


MINTYPE_SINGLETON = MinType()


# noinspection PyPep8Naming
class attrgetter_nonesort:
    """
    Modification of ``operator.attrgetter``.
    Returns an object's attributes, or the ``mintype_singleton`` if the
    attribute is ``None``.
    """
    __slots__ = ('_attrs', '_call')

    def __init__(self, attr, *attrs):
        if not attrs:
            if not isinstance(attr, str):
                raise TypeError('attribute name must be a string')
            self._attrs = (attr,)
            names = attr.split('.')

            def func(obj):
                for name in names:
                    obj = getattr(obj, name)
                if obj is None:  # MODIFIED HERE
                    return MINTYPE_SINGLETON
                return obj

            self._call = func
        else:
            self._attrs = (attr,) + attrs
            # MODIFIED HERE:
            getters = tuple(map(attrgetter_nonesort, self._attrs))

            def func(obj):
                return tuple(getter(obj) for getter in getters)

            self._call = func

    def __call__(self, obj):
        return self._call(obj)

    def __repr__(self):
        return '%s.%s(%s)' % (self.__class__.__module__,
                              self.__class__.__qualname__,
                              ', '.join(map(repr, self._attrs)))

    def __reduce__(self):
        return self.__class__, self._attrs


# noinspection PyPep8Naming
class methodcaller_nonesort:
    """
    As per :class:`attrgetter_nonesort` (q.v.), but for ``methodcaller``.
    """
    __slots__ = ('_name', '_args', '_kwargs')

    def __init__(*args, **kwargs):
        if len(args) < 2:
            msg = "methodcaller needs at least one argument, the method name"
            raise TypeError(msg)
        self = args[0]
        self._name = args[1]
        if not isinstance(self._name, str):
            raise TypeError('method name must be a string')
        self._args = args[2:]
        self._kwargs = kwargs

    def __call__(self, obj):
        # MODIFICATION HERE
        result = getattr(obj, self._name)(*self._args, **self._kwargs)
        if result is None:
            return MINTYPE_SINGLETON
        return result

    def __repr__(self):
        args = [repr(self._name)]
        args.extend(map(repr, self._args))
        args.extend('%s=%r' % (k, v) for k, v in self._kwargs.items())
        return '%s.%s(%s)' % (self.__class__.__module__,
                              self.__class__.__name__,
                              ', '.join(args))

    def __reduce__(self):
        if not self._kwargs:
            return self.__class__, (self._name,) + self._args
        else:
            return (
                partial(self.__class__, self._name, **self._kwargs),
                self._args
            )



#!/usr/bin/env python
# cardinal_pythonlib/sqlalchemy/schema.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Functions to work with SQLAlchemy schemas (schemata) directly, via SQLAlchemy
Core.**

"""

import ast
import contextlib
import copy
import csv
from functools import lru_cache
import io
import re
from typing import Any, Dict, Generator, List, Optional, Type, Union

from sqlalchemy.dialects import mssql, mysql
# noinspection PyProtectedMember
from sqlalchemy.engine import Connection, Engine, ResultProxy
from sqlalchemy.engine.interfaces import Dialect
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.dialects.mssql.base import TIMESTAMP as MSSQL_TIMESTAMP
from sqlalchemy.schema import (Column, CreateColumn, DDL, MetaData, Index,
                               Sequence, Table)
from sqlalchemy.sql import sqltypes, text
from sqlalchemy.sql.sqltypes import BigInteger, TypeEngine
from sqlalchemy.sql.visitors import VisitableType

from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler
from cardinal_pythonlib.sqlalchemy.dialect import (
    quote_identifier,
    SqlaDialectName,
)

log = get_brace_style_log_with_null_handler(__name__)

# =============================================================================
# Constants
# =============================================================================

MSSQL_DEFAULT_SCHEMA = 'dbo'
POSTGRES_DEFAULT_SCHEMA = 'public'


# =============================================================================
# Inspect tables (SQLAlchemy Core)
# =============================================================================

def get_table_names(engine: Engine) -> List[str]:
    """
    Returns a list of database table names from the :class:`Engine`.
    """
    insp = Inspector.from_engine(engine)
    return insp.get_table_names()


def get_view_names(engine: Engine) -> List[str]:
    """
    Returns a list of database view names from the :class:`Engine`.
    """
    insp = Inspector.from_engine(engine)
    return insp.get_view_names()


def table_exists(engine: Engine, tablename: str) -> bool:
    """
    Does the named table exist in the database?
    """
    return tablename in get_table_names(engine)


def view_exists(engine: Engine, viewname: str) -> bool:
    """
    Does the named view exist in the database?
    """
    return viewname in get_view_names(engine)


def table_or_view_exists(engine: Engine, table_or_view_name: str) -> bool:
    """
    Does the named table/view exist (either as a table or as a view) in the
    database?
    """
    tables_and_views = get_table_names(engine) + get_view_names(engine)
    return table_or_view_name in tables_and_views


class SqlaColumnInspectionInfo(object):
    """
    Class to represent information from inspecting a database column.

    A clearer way of getting information than the plain ``dict`` that SQLAlchemy
    uses.
    """
    def __init__(self, sqla_info_dict: Dict[str, Any]) -> None:
        """
        Args:
            sqla_info_dict:
                see
        
                - https://docs.sqlalchemy.org/en/latest/core/reflection.html#sqlalchemy.engine.reflection.Inspector.get_columns
                - https://bitbucket.org/zzzeek/sqlalchemy/issues/4051/sqlalchemyenginereflectioninspectorget_col
        """  # noqa
        # log.debug(repr(sqla_info_dict))
        self.name = sqla_info_dict['name']  # type: str
        self.type = sqla_info_dict['type']  # type: TypeEngine
        self.nullable = sqla_info_dict['nullable']  # type: bool
        self.default = sqla_info_dict['default']  # type: str  # SQL string expression  # noqa
        self.attrs = sqla_info_dict.get('attrs', {})  # type: Dict[str, Any]
        self.comment = sqla_info_dict.get('comment', '')
        # ... NB not appearing in


def gen_columns_info(engine: Engine,
                     tablename: str) -> Generator[SqlaColumnInspectionInfo,
                                                  None, None]:
    """
    For the specified table, generate column information as
    :class:`SqlaColumnInspectionInfo` objects.
    """
    # Dictionary structure: see
    # http://docs.sqlalchemy.org/en/latest/core/reflection.html#sqlalchemy.engine.reflection.Inspector.get_columns  # noqa
    insp = Inspector.from_engine(engine)
    for d in insp.get_columns(tablename):
        yield SqlaColumnInspectionInfo(d)


def get_column_info(engine: Engine, tablename: str,
                    columnname: str) -> Optional[SqlaColumnInspectionInfo]:
    """
    For the specified column in the specified table, get column information
    as a :class:`SqlaColumnInspectionInfo` object (or ``None`` if such a
    column can't be found).
    """
    for info in gen_columns_info(engine, tablename):
        if info.name == columnname:
            return info
    return None


def get_column_type(engine: Engine, tablename: str,
                    columnname: str) -> Optional[TypeEngine]:
    """
    For the specified column in the specified table, get its type as an
    instance of an SQLAlchemy column type class (or ``None`` if such a column
    can't be found).

    For more on :class:`TypeEngine`, see
    :func:`cardinal_pythonlib.orm_inspect.coltype_as_typeengine`.
    """
    for info in gen_columns_info(engine, tablename):
        if info.name == columnname:
            return info.type
    return None


def get_column_names(engine: Engine, tablename: str) -> List[str]:
    """
    Get all the database column names for the specified table.
    """
    return [info.name for info in gen_columns_info(engine, tablename)]


# =============================================================================
# More introspection
# =============================================================================

def get_pk_colnames(table_: Table) -> List[str]:
    """
    If a table has a PK, this will return its database column name(s);
    otherwise, ``None``.
    """
    pk_names = []  # type: List[str]
    for col in table_.columns:
        if col.primary_key:
            pk_names.append(col.name)
    return pk_names


def get_single_int_pk_colname(table_: Table) -> Optional[str]:
    """
    If a table has a single-field (non-composite) integer PK, this will
    return its database column name; otherwise, None.

    Note that it is legitimate for a database table to have both a composite
    primary key and a separate ``IDENTITY`` (``AUTOINCREMENT``) integer field.
    This function won't find such columns.
    """
    n_pks = 0
    int_pk_names = []
    for col in table_.columns:
        if col.primary_key:
            n_pks += 1
            if is_sqlatype_integer(col.type):
                int_pk_names.append(col.name)
    if n_pks == 1 and len(int_pk_names) == 1:
        return int_pk_names[0]
    return None


def get_single_int_autoincrement_colname(table_: Table) -> Optional[str]:
    """
    If a table has a single integer ``AUTOINCREMENT`` column, this will
    return its name; otherwise, ``None``.

    - It's unlikely that a database has >1 ``AUTOINCREMENT`` field anyway, but
      we should check.
    - SQL Server's ``IDENTITY`` keyword is equivalent to MySQL's
      ``AUTOINCREMENT``.
    - Verify against SQL Server:

      .. code-block:: sql

        SELECT table_name, column_name
        FROM information_schema.columns
        WHERE COLUMNPROPERTY(OBJECT_ID(table_schema + '.' + table_name),
                             column_name,
                             'IsIdentity') = 1
        ORDER BY table_name;

      ... https://stackoverflow.com/questions/87747

    - Also:

      .. code-block:: sql

        sp_columns 'tablename';

      ... which is what SQLAlchemy does (``dialects/mssql/base.py``, in
      :func:`get_columns`).
    """
    n_autoinc = 0
    int_autoinc_names = []
    for col in table_.columns:
        if col.autoincrement:
            n_autoinc += 1
            if is_sqlatype_integer(col.type):
                int_autoinc_names.append(col.name)
    if n_autoinc > 1:
        log.warning("Table {!r} has {} autoincrement columns",
                    table_.name, n_autoinc)
    if n_autoinc == 1 and len(int_autoinc_names) == 1:
        return int_autoinc_names[0]
    return None


def get_effective_int_pk_col(table_: Table) -> Optional[str]:
    """
    If a table has a single integer primary key, or a single integer
    ``AUTOINCREMENT`` column, return its column name; otherwise, ``None``.
    """
    return (
        get_single_int_pk_colname(table_) or
        get_single_int_autoincrement_colname(table_) or
        None
    )


# =============================================================================
# Indexes
# =============================================================================

def index_exists(engine: Engine, tablename: str, indexname: str) -> bool:
    """
    Does the specified index exist for the specified table?
    """
    insp = Inspector.from_engine(engine)
    return any(i['name'] == indexname for i in insp.get_indexes(tablename))


def mssql_get_pk_index_name(engine: Engine,
                            tablename: str,
                            schemaname: str = MSSQL_DEFAULT_SCHEMA) -> str:
    """
    For Microsoft SQL Server specifically: fetch the name of the PK index
    for the specified table (in the specified schema), or ``''`` if none is
    found.
    """
    # http://docs.sqlalchemy.org/en/latest/core/connections.html#sqlalchemy.engine.Connection.execute  # noqa
    # http://docs.sqlalchemy.org/en/latest/core/sqlelement.html#sqlalchemy.sql.expression.text  # noqa
    # http://docs.sqlalchemy.org/en/latest/core/sqlelement.html#sqlalchemy.sql.expression.TextClause.bindparams  # noqa
    # http://docs.sqlalchemy.org/en/latest/core/connections.html#sqlalchemy.engine.ResultProxy  # noqa
    query = text("""
SELECT
    kc.name AS index_name
FROM
    sys.key_constraints AS kc
    INNER JOIN sys.tables AS ta ON ta.object_id = kc.parent_object_id
    INNER JOIN sys.schemas AS s ON ta.schema_id = s.schema_id
WHERE
    kc.[type] = 'PK'
    AND ta.name = :tablename
    AND s.name = :schemaname
    """).bindparams(
        tablename=tablename,
        schemaname=schemaname,
    )
    with contextlib.closing(
            engine.execute(query)) as result:  # type: ResultProxy  # noqa
        row = result.fetchone()
        return row[0] if row else ''


def mssql_table_has_ft_index(engine: Engine,
                             tablename: str,
                             schemaname: str = MSSQL_DEFAULT_SCHEMA) -> bool:
    """
    For Microsoft SQL Server specifically: does the specified table (in the
    specified schema) have at least one full-text index?
    """
    query = text("""
SELECT
    COUNT(*)
FROM
    sys.key_constraints AS kc
    INNER JOIN sys.tables AS ta ON ta.object_id = kc.parent_object_id
    INNER JOIN sys.schemas AS s ON ta.schema_id = s.schema_id
    INNER JOIN sys.fulltext_indexes AS fi ON fi.object_id = ta.object_id
WHERE
    ta.name = :tablename
    AND s.name = :schemaname
    """).bindparams(
        tablename=tablename,
        schemaname=schemaname,
    )
    with contextlib.closing(
            engine.execute(query)) as result:  # type: ResultProxy  # noqa
        row = result.fetchone()
        return row[0] > 0


def mssql_transaction_count(engine_or_conn: Union[Connection, Engine]) -> int:
    """
    For Microsoft SQL Server specifically: fetch the value of the ``TRANCOUNT``
    variable (see e.g.
    https://docs.microsoft.com/en-us/sql/t-sql/functions/trancount-transact-sql?view=sql-server-2017).
    Returns ``None`` if it can't be found (unlikely?).
    """
    sql = "SELECT @@TRANCOUNT"
    with contextlib.closing(
            engine_or_conn.execute(sql)) as result:  # type: ResultProxy  # noqa
        row = result.fetchone()
        return row[0] if row else None


def add_index(engine: Engine,
              sqla_column: Column = None,
              multiple_sqla_columns: List[Column] = None,
              unique: bool = False,
              fulltext: bool = False,
              length: int = None) -> None:
    """
    Adds an index to a database column (or, in restricted circumstances,
    several columns).

    The table name is worked out from the :class:`Column` object.

    Args:
        engine: SQLAlchemy :class:`Engine` object
        sqla_column: single column to index
        multiple_sqla_columns: multiple columns to index (see below)
        unique: make a ``UNIQUE`` index?
        fulltext: make a ``FULLTEXT`` index?
        length: index length to use (default ``None``)

    Restrictions:

    - Specify either ``sqla_column`` or ``multiple_sqla_columns``, not both.
    - The normal method is ``sqla_column``.
    - ``multiple_sqla_columns`` is only used for Microsoft SQL Server full-text
      indexing (as this database permits only one full-text index per table,
      though that index can be on multiple columns).

    """
    # We used to process a table as a unit; this makes index creation faster
    # (using ALTER TABLE).
    # http://dev.mysql.com/doc/innodb/1.1/en/innodb-create-index-examples.html  # noqa
    # ... ignored in transition to SQLAlchemy

    def quote(identifier: str) -> str:
        return quote_identifier(identifier, engine)

    is_mssql = engine.dialect.name == SqlaDialectName.MSSQL
    is_mysql = engine.dialect.name == SqlaDialectName.MYSQL

    multiple_sqla_columns = multiple_sqla_columns or []  # type: List[Column]
    if multiple_sqla_columns and not (fulltext and is_mssql):
        raise ValueError("add_index: Use multiple_sqla_columns only for mssql "
                         "(Microsoft SQL Server) full-text indexing")
    if bool(multiple_sqla_columns) == (sqla_column is not None):
        raise ValueError(
            f"add_index: Use either sqla_column or multiple_sqla_columns, "
            f"not both (sqla_column = {sqla_column!r}, "
            f"multiple_sqla_columns = {multiple_sqla_columns!r}"
        )
    if sqla_column is not None:
        colnames = [sqla_column.name]
        sqla_table = sqla_column.table
        tablename = sqla_table.name
    else:
        colnames = [c.name for c in multiple_sqla_columns]
        sqla_table = multiple_sqla_columns[0].table
        tablename = sqla_table.name
        if any(c.table.name != tablename for c in multiple_sqla_columns[1:]):
            raise ValueError(
                f"add_index: tablenames are inconsistent in "
                f"multiple_sqla_columns = {multiple_sqla_columns!r}")

    if fulltext:
        if is_mssql:
            idxname = ''  # they are unnamed
        else:
            idxname = "_idxft_{}".format("_".join(colnames))
    else:
        idxname = "_idx_{}".format("_".join(colnames))
    if idxname and index_exists(engine, tablename, idxname):
        log.info(f"Skipping creation of index {idxname} on "
                 f"table {tablename}; already exists")
        return
        # because it will crash if you add it again!
    log.info(
        "Creating{ft} index {i} on table {t}, column(s) {c}",
        ft=" full-text" if fulltext else "",
        i=idxname or "<unnamed>",
        t=tablename,
        c=", ".join(colnames),
    )

    if fulltext:
        if is_mysql:
            log.info('OK to ignore this warning, if it follows next: '
                     '"InnoDB rebuilding table to add column FTS_DOC_ID"')
            # https://dev.mysql.com/doc/refman/5.6/en/innodb-fulltext-index.html
            sql = (
                "ALTER TABLE {tablename} "
                "ADD FULLTEXT INDEX {idxname} ({colnames})".format(
                    tablename=quote(tablename),
                    idxname=quote(idxname),
                    colnames=", ".join(quote(c) for c in colnames),
                )
            )
            # DDL(sql, bind=engine).execute_if(dialect=SqlaDialectName.MYSQL)
            DDL(sql, bind=engine).execute()

        elif is_mssql:  # Microsoft SQL Server
            # https://msdn.microsoft.com/library/ms187317(SQL.130).aspx
            # Argh! Complex.
            # Note that the database must also have had a
            #   CREATE FULLTEXT CATALOG somename AS DEFAULT;
            # statement executed on it beforehand.
            schemaname = engine.schema_for_object(
                sqla_table) or MSSQL_DEFAULT_SCHEMA  # noqa
            if mssql_table_has_ft_index(engine=engine,
                                        tablename=tablename,
                                        schemaname=schemaname):
                log.info(
                    f"... skipping creation of full-text index on table "
                    f"{tablename}; a full-text index already exists for that "
                    f"table; you can have only one full-text index per table, "
                    f"though it can be on multiple columns")
                return
            pk_index_name = mssql_get_pk_index_name(
                engine=engine, tablename=tablename, schemaname=schemaname)
            if not pk_index_name:
                raise ValueError(
                    f"To make a FULLTEXT index under SQL Server, we need to "
                    f"know the name of the PK index, but couldn't find one "
                    f"from get_pk_index_name() for table {tablename!r}")
            # We don't name the FULLTEXT index itself, but it has to relate
            # to an existing unique index.
            sql = (
                "CREATE FULLTEXT INDEX ON {tablename} ({colnames}) "
                "KEY INDEX {keyidxname} ".format(
                    tablename=quote(tablename),
                    keyidxname=quote(pk_index_name),
                    colnames=", ".join(quote(c) for c in colnames),
                )
            )
            # SQL Server won't let you do this inside a transaction:
            # "CREATE FULLTEXT INDEX statement cannot be used inside a user
            # transaction."
            # https://msdn.microsoft.com/nl-nl/library/ms191544(v=sql.105).aspx
            # So let's ensure any preceding transactions are completed, and
            # run the SQL in a raw way:
            # engine.execute(sql).execution_options(autocommit=False)
            # http://docs.sqlalchemy.org/en/latest/core/connections.html#understanding-autocommit
            #
            # ... lots of faff with this (see test code in no_transactions.py)
            # ... ended up using explicit "autocommit=True" parameter (for
            #     pyodbc); see create_indexes()
            transaction_count = mssql_transaction_count(engine)
            if transaction_count != 0:
                log.critical(f"SQL Server transaction count (should be 0): "
                             f"{transaction_count}")
                # Executing serial COMMITs or a ROLLBACK won't help here if
                # this transaction is due to Python DBAPI default behaviour.
            DDL(sql, bind=engine).execute()

            # The reversal procedure is DROP FULLTEXT INDEX ON tablename;

        else:
            log.error(f"Don't know how to make full text index on dialect "
                      f"{engine.dialect.name}")

    else:
        index = Index(idxname, sqla_column, unique=unique, mysql_length=length)
        index.create(engine)
        # Index creation doesn't require a commit.


# =============================================================================
# More DDL
# =============================================================================

def make_bigint_autoincrement_column(column_name: str,
                                     dialect: Dialect) -> Column:
    """
    Returns an instance of :class:`Column` representing a :class:`BigInteger`
    ``AUTOINCREMENT`` column in the specified :class:`Dialect`.
    """
    # noinspection PyUnresolvedReferences
    if dialect.name == SqlaDialectName.MSSQL:
        return Column(column_name, BigInteger,
                      Sequence('dummy_name', start=1, increment=1))
    else:
        # return Column(column_name, BigInteger, autoincrement=True)
        # noinspection PyUnresolvedReferences
        raise AssertionError(
            f"SQLAlchemy doesn't support non-PK autoincrement fields yet for "
            f"dialect {dialect.name!r}")
        # see https://stackoverflow.com/questions/2937229


def column_creation_ddl(sqla_column: Column, dialect: Dialect) -> str:
    """
    Returns DDL to create a column, using the specified dialect.

    The column should already be bound to a table (because e.g. the SQL Server
    dialect requires this for DDL generation).

    Manual testing:
    
    .. code-block:: python

        from sqlalchemy.schema import Column, CreateColumn, MetaData, Sequence, Table
        from sqlalchemy.sql.sqltypes import BigInteger
        from sqlalchemy.dialects.mssql.base import MSDialect
        dialect = MSDialect()
        col1 = Column('hello', BigInteger, nullable=True)
        col2 = Column('world', BigInteger, autoincrement=True)  # does NOT generate IDENTITY
        col3 = Column('you', BigInteger, Sequence('dummy_name', start=1, increment=1))
        metadata = MetaData()
        t = Table('mytable', metadata)
        t.append_column(col1)
        t.append_column(col2)
        t.append_column(col3)
        print(str(CreateColumn(col1).compile(dialect=dialect)))  # hello BIGINT NULL
        print(str(CreateColumn(col2).compile(dialect=dialect)))  # world BIGINT NULL
        print(str(CreateColumn(col3).compile(dialect=dialect)))  # you BIGINT NOT NULL IDENTITY(1,1)

    If you don't append the column to a Table object, the DDL generation step
    gives:
    
    .. code-block:: none

        sqlalchemy.exc.CompileError: mssql requires Table-bound columns in order to generate DDL
    """  # noqa
    return str(CreateColumn(sqla_column).compile(dialect=dialect))


# noinspection PyUnresolvedReferences
def giant_text_sqltype(dialect: Dialect) -> str:
    """
    Returns the SQL column type used to make very large text columns for a
    given dialect.

    Args:
        dialect: a SQLAlchemy :class:`Dialect`
    Returns:
        the SQL data type of "giant text", typically 'LONGTEXT' for MySQL
        and 'NVARCHAR(MAX)' for SQL Server.
    """
    if dialect.name == SqlaDialectName.SQLSERVER:
        return 'NVARCHAR(MAX)'
    elif dialect.name == SqlaDialectName.MYSQL:
        return 'LONGTEXT'
    else:
        raise ValueError(f"Unknown dialect: {dialect.name}")


# =============================================================================
# SQLAlchemy column types
# =============================================================================

# -----------------------------------------------------------------------------
# Reverse a textual SQL column type to an SQLAlchemy column type
# -----------------------------------------------------------------------------

RE_MYSQL_ENUM_COLTYPE = re.compile(r'ENUM\((?P<valuelist>.+)\)')
RE_COLTYPE_WITH_COLLATE = re.compile(r'(?P<maintype>.+) COLLATE .+')
RE_COLTYPE_WITH_ONE_PARAM = re.compile(r'(?P<type>\w+)\((?P<size>\w+)\)')
# ... e.g. "VARCHAR(10)"
RE_COLTYPE_WITH_TWO_PARAMS = re.compile(
    r'(?P<type>\w+)\((?P<size>\w+),\s*(?P<dp>\w+)\)')
# ... e.g. "DECIMAL(10, 2)"


# http://www.w3schools.com/sql/sql_create_table.asp


def _get_sqla_coltype_class_from_str(coltype: str,
                                     dialect: Dialect) -> Type[TypeEngine]:
    """
    Returns the SQLAlchemy class corresponding to a particular SQL column
    type in a given dialect.

    Performs an upper- and lower-case search.
    For example, the SQLite dialect uses upper case, and the
    MySQL dialect uses lower case.
    """
    # noinspection PyUnresolvedReferences
    ischema_names = dialect.ischema_names
    try:
        return ischema_names[coltype.upper()]
    except KeyError:
        return ischema_names[coltype.lower()]


def get_list_of_sql_string_literals_from_quoted_csv(x: str) -> List[str]:
    """
    Used to extract SQL column type parameters. For example, MySQL has column
    types that look like ``ENUM('a', 'b', 'c', 'd')``. This function takes the
    ``"'a', 'b', 'c', 'd'"`` and converts it to ``['a', 'b', 'c', 'd']``.
    """
    f = io.StringIO(x)
    reader = csv.reader(f, delimiter=',', quotechar="'", quoting=csv.QUOTE_ALL,
                        skipinitialspace=True)
    for line in reader:  # should only be one
        return [x for x in line]


@lru_cache(maxsize=None)
def get_sqla_coltype_from_dialect_str(coltype: str,
                                      dialect: Dialect) -> TypeEngine:
    """
    Returns an SQLAlchemy column type, given a column type name (a string) and
    an SQLAlchemy dialect. For example, this might convert the string
    ``INTEGER(11)`` to an SQLAlchemy ``Integer(length=11)``.

    NOTE that the reverse operation is performed by ``str(coltype)`` or
    ``coltype.compile()`` or ``coltype.compile(dialect)``; see
    :class:`TypeEngine`.

    Args:
        dialect: a SQLAlchemy :class:`Dialect` class

        coltype: a ``str()`` representation, e.g. from ``str(c['type'])`` where
            ``c`` is an instance of :class:`sqlalchemy.sql.schema.Column`.

    Returns:
        a Python object that is a subclass of
        :class:`sqlalchemy.types.TypeEngine`

    Example:

    .. code-block:: python

        get_sqla_coltype_from_string('INTEGER(11)', engine.dialect)
        # gives: Integer(length=11)

    Notes:

    - :class:`sqlalchemy.engine.default.DefaultDialect` is the dialect base
      class

    - a dialect contains these things of interest:

      - ``ischema_names``: string-to-class dictionary
      - ``type_compiler``: instance of e.g.
        :class:`sqlalchemy.sql.compiler.GenericTypeCompiler`. This has a
        ``process()`` method, but that operates on :class:`TypeEngine` objects.
      - ``get_columns``: takes a table name, inspects the database

    - example of the dangers of ``eval``:
      https://nedbatchelder.com/blog/201206/eval_really_is_dangerous.html

    - An example of a function doing the reflection/inspection within
      SQLAlchemy is
      :func:`sqlalchemy.dialects.mssql.base.MSDialect.get_columns`,
      which has this lookup: ``coltype = self.ischema_names.get(type, None)``

    Caveats:

    - the parameters, e.g. ``DATETIME(6)``, do NOT necessarily either work at
      all or work correctly. For example, SQLAlchemy will happily spit out
      ``'INTEGER(11)'`` but its :class:`sqlalchemy.sql.sqltypes.INTEGER` class
      takes no parameters, so you get the error ``TypeError: object() takes no
      parameters``. Similarly, MySQL's ``DATETIME(6)`` uses the 6 to refer to
      precision, but the ``DATETIME`` class in SQLAlchemy takes only a boolean
      parameter (timezone).
    - However, sometimes we have to have parameters, e.g. ``VARCHAR`` length.
    - Thus, this is a bit useless.
    - Fixed, with a few special cases.
    """
    size = None  # type: Optional[int]
    dp = None  # type: Optional[int]
    args = []  # type: List[Any]
    kwargs = {}  # type: Dict[str, Any]
    basetype = ''

    # noinspection PyPep8,PyBroadException
    try:
        # Split e.g. "VARCHAR(32) COLLATE blah" into "VARCHAR(32)", "who cares"
        m = RE_COLTYPE_WITH_COLLATE.match(coltype)
        if m is not None:
            coltype = m.group('maintype')

        found = False

        if not found:
            # Deal with ENUM('a', 'b', 'c', ...)
            m = RE_MYSQL_ENUM_COLTYPE.match(coltype)
            if m is not None:
                # Convert to VARCHAR with max size being that of largest enum
                basetype = 'VARCHAR'
                values = get_list_of_sql_string_literals_from_quoted_csv(
                    m.group('valuelist'))
                length = max(len(x) for x in values)
                kwargs = {'length': length}
                found = True

        if not found:
            # Split e.g. "DECIMAL(10, 2)" into DECIMAL, 10, 2
            m = RE_COLTYPE_WITH_TWO_PARAMS.match(coltype)
            if m is not None:
                basetype = m.group('type').upper()
                size = ast.literal_eval(m.group('size'))
                dp = ast.literal_eval(m.group('dp'))
                found = True

        if not found:
            # Split e.g. "VARCHAR(32)" into VARCHAR, 32
            m = RE_COLTYPE_WITH_ONE_PARAM.match(coltype)
            if m is not None:
                basetype = m.group('type').upper()
                size_text = m.group('size').strip().upper()
                if size_text != 'MAX':
                    size = ast.literal_eval(size_text)
                found = True

        if not found:
            basetype = coltype.upper()

        # Special cases: pre-processing
        # noinspection PyUnresolvedReferences
        if (dialect.name == SqlaDialectName.MSSQL and
                basetype.lower() == 'integer'):
            basetype = 'int'

        cls = _get_sqla_coltype_class_from_str(basetype, dialect)

        # Special cases: post-processing
        if basetype == 'DATETIME' and size:
            # First argument to DATETIME() is timezone, so...
            # noinspection PyUnresolvedReferences
            if dialect.name == SqlaDialectName.MYSQL:
                kwargs = {'fsp': size}
            else:
                pass
        else:
            args = [x for x in (size, dp) if x is not None]

        try:
            return cls(*args, **kwargs)
        except TypeError:
            return cls()

    except:
        # noinspection PyUnresolvedReferences
        raise ValueError(f"Failed to convert SQL type {coltype!r} in dialect "
                         f"{dialect.name!r} to an SQLAlchemy type")


# get_sqla_coltype_from_dialect_str("INTEGER", engine.dialect)
# get_sqla_coltype_from_dialect_str("INTEGER(11)", engine.dialect)
# get_sqla_coltype_from_dialect_str("VARCHAR(50)", engine.dialect)
# get_sqla_coltype_from_dialect_str("DATETIME", engine.dialect)
# get_sqla_coltype_from_dialect_str("DATETIME(6)", engine.dialect)


# =============================================================================
# Do special dialect conversions on SQLAlchemy SQL types (of class type)
# =============================================================================

def remove_collation(coltype: TypeEngine) -> TypeEngine:
    """
    Returns a copy of the specific column type with any ``COLLATION`` removed.
    """
    if not getattr(coltype, 'collation', None):
        return coltype
    newcoltype = copy.copy(coltype)
    newcoltype.collation = None
    return newcoltype


@lru_cache(maxsize=None)
def convert_sqla_type_for_dialect(
        coltype: TypeEngine,
        dialect: Dialect,
        strip_collation: bool = True,
        convert_mssql_timestamp: bool = True,
        expand_for_scrubbing: bool = False) -> TypeEngine:
    """
    Converts an SQLAlchemy column type from one SQL dialect to another.

    Args:
        coltype: SQLAlchemy column type in the source dialect

        dialect: destination :class:`Dialect`

        strip_collation: remove any ``COLLATION`` information?

        convert_mssql_timestamp:
            since you cannot write to a SQL Server ``TIMESTAMP`` field, setting
            this option to ``True`` (the default) converts such types to
            something equivalent but writable.

        expand_for_scrubbing:
            The purpose of expand_for_scrubbing is that, for example, a
            ``VARCHAR(200)`` field containing one or more instances of
            ``Jones``, where ``Jones`` is to be replaced with ``[XXXXXX]``,
            will get longer (by an unpredictable amount). So, better to expand
            to unlimited length.

    Returns:
        an SQLAlchemy column type instance, in the destination dialect

    """
    assert coltype is not None

    # noinspection PyUnresolvedReferences
    to_mysql = dialect.name == SqlaDialectName.MYSQL
    # noinspection PyUnresolvedReferences
    to_mssql = dialect.name == SqlaDialectName.MSSQL
    typeclass = type(coltype)

    # -------------------------------------------------------------------------
    # Text
    # -------------------------------------------------------------------------
    if isinstance(coltype, sqltypes.Enum):
        return sqltypes.String(length=coltype.length)
    if isinstance(coltype, sqltypes.UnicodeText):
        # Unbounded Unicode text.
        # Includes derived classes such as mssql.base.NTEXT.
        return sqltypes.UnicodeText()
    if isinstance(coltype, sqltypes.Text):
        # Unbounded text, more generally. (UnicodeText inherits from Text.)
        # Includes sqltypes.TEXT.
        return sqltypes.Text()
    # Everything inheriting from String has a length property, but can be None.
    # There are types that can be unlimited in SQL Server, e.g. VARCHAR(MAX)
    # and NVARCHAR(MAX), that MySQL needs a length for. (Failure to convert
    # gives e.g.: 'NVARCHAR requires a length on dialect mysql'.)
    if isinstance(coltype, sqltypes.Unicode):
        # Includes NVARCHAR(MAX) in SQL -> NVARCHAR() in SQLAlchemy.
        if (coltype.length is None and to_mysql) or expand_for_scrubbing:
            return sqltypes.UnicodeText()
    # The most general case; will pick up any other string types.
    if isinstance(coltype, sqltypes.String):
        # Includes VARCHAR(MAX) in SQL -> VARCHAR() in SQLAlchemy
        if (coltype.length is None and to_mysql) or expand_for_scrubbing:
            return sqltypes.Text()
        if strip_collation:
            return remove_collation(coltype)
        return coltype

    # -------------------------------------------------------------------------
    # Binary
    # -------------------------------------------------------------------------

    # -------------------------------------------------------------------------
    # BIT
    # -------------------------------------------------------------------------
    if typeclass == mssql.base.BIT and to_mysql:
        # MySQL BIT objects have a length attribute.
        return mysql.base.BIT()

    # -------------------------------------------------------------------------
    # TIMESTAMP
    # -------------------------------------------------------------------------
    is_mssql_timestamp = isinstance(coltype, MSSQL_TIMESTAMP)
    if is_mssql_timestamp and to_mssql and convert_mssql_timestamp:
        # You cannot write explicitly to a TIMESTAMP field in SQL Server; it's
        # used for autogenerated values only.
        # - https://stackoverflow.com/questions/10262426/sql-server-cannot-insert-an-explicit-value-into-a-timestamp-column  # noqa
        # - https://social.msdn.microsoft.com/Forums/sqlserver/en-US/5167204b-ef32-4662-8e01-00c9f0f362c2/how-to-tranfer-a-column-with-timestamp-datatype?forum=transactsql  # noqa
        #   ... suggesting BINARY(8) to store the value.
        # MySQL is more helpful:
        # - https://stackoverflow.com/questions/409286/should-i-use-field-datetime-or-timestamp  # noqa
        return mssql.base.BINARY(8)

    # -------------------------------------------------------------------------
    # Some other type
    # -------------------------------------------------------------------------
    return coltype


# =============================================================================
# Questions about SQLAlchemy column types
# =============================================================================

# Note:
#   x = String        } type(x) == VisitableType  # metaclass
#   x = BigInteger    }
# but:
#   x = String()      } type(x) == TypeEngine
#   x = BigInteger()  }
#
# isinstance also cheerfully handles multiple inheritance, i.e. if you have
# class A(object), class B(object), and class C(A, B), followed by x = C(),
# then all of isinstance(x, A), isinstance(x, B), isinstance(x, C) are True

def _coltype_to_typeengine(coltype: Union[TypeEngine,
                                          VisitableType]) -> TypeEngine:
    """
    An example is simplest: if you pass in ``Integer()`` (an instance of
    :class:`TypeEngine`), you'll get ``Integer()`` back. If you pass in
    ``Integer`` (an instance of :class:`VisitableType`), you'll also get
    ``Integer()`` back. The function asserts that its return type is an
    instance of :class:`TypeEngine`.

    See also
    :func:`cardinal_pythonlib.sqlalchemy.orm_inspect.coltype_as_typeengine`.
    """
    if isinstance(coltype, VisitableType):
        coltype = coltype()
    assert isinstance(coltype, TypeEngine)
    return coltype


def is_sqlatype_binary(coltype: Union[TypeEngine, VisitableType]) -> bool:
    """
    Is the SQLAlchemy column type a binary type?
    """
    # Several binary types inherit internally from _Binary, making that the
    # easiest to check.
    coltype = _coltype_to_typeengine(coltype)
    # noinspection PyProtectedMember
    return isinstance(coltype, sqltypes._Binary)


def is_sqlatype_date(coltype: Union[TypeEngine, VisitableType]) -> bool:
    """
    Is the SQLAlchemy column type a date type?
    """
    coltype = _coltype_to_typeengine(coltype)
    # No longer valid in SQLAlchemy 1.2.11:
    # return isinstance(coltype, sqltypes._DateAffinity)
    return (
        isinstance(coltype, sqltypes.DateTime) or
        isinstance(coltype, sqltypes.Date)
    )


def is_sqlatype_integer(coltype: Union[TypeEngine, VisitableType]) -> bool:
    """
    Is the SQLAlchemy column type an integer type?
    """
    coltype = _coltype_to_typeengine(coltype)
    return isinstance(coltype, sqltypes.Integer)


def is_sqlatype_numeric(coltype: Union[TypeEngine, VisitableType]) -> bool:
    """
    Is the SQLAlchemy column type one that inherits from :class:`Numeric`,
    such as :class:`Float`, :class:`Decimal`?
    """
    coltype = _coltype_to_typeengine(coltype)
    return isinstance(coltype, sqltypes.Numeric)  # includes Float, Decimal


def is_sqlatype_string(coltype: Union[TypeEngine, VisitableType]) -> bool:
    """
    Is the SQLAlchemy column type a string type?
    """
    coltype = _coltype_to_typeengine(coltype)
    return isinstance(coltype, sqltypes.String)


def is_sqlatype_text_of_length_at_least(
        coltype: Union[TypeEngine, VisitableType],
        min_length: int = 1000) -> bool:
    """
    Is the SQLAlchemy column type a string type that's at least the specified
    length?
    """
    coltype = _coltype_to_typeengine(coltype)
    if not isinstance(coltype, sqltypes.String):
        return False  # not a string/text type at all
    if coltype.length is None:
        return True  # string of unlimited length
    return coltype.length >= min_length


def is_sqlatype_text_over_one_char(
        coltype: Union[TypeEngine, VisitableType]) -> bool:
    """
    Is the SQLAlchemy column type a string type that's more than one character
    long?
    """
    coltype = _coltype_to_typeengine(coltype)
    return is_sqlatype_text_of_length_at_least(coltype, 2)


def does_sqlatype_merit_fulltext_index(
        coltype: Union[TypeEngine, VisitableType],
        min_length: int = 1000) -> bool:
    """
    Is the SQLAlchemy column type a type that might merit a ``FULLTEXT``
    index (meaning a string type of at least ``min_length``)?
    """
    coltype = _coltype_to_typeengine(coltype)
    return is_sqlatype_text_of_length_at_least(coltype, min_length)


def does_sqlatype_require_index_len(
        coltype: Union[TypeEngine, VisitableType]) -> bool:
    """
    Is the SQLAlchemy column type one that requires its indexes to have a
    length specified?

    (MySQL, at least, requires index length to be specified for ``BLOB`` and
    ``TEXT`` columns:
    https://dev.mysql.com/doc/refman/5.7/en/create-index.html.)
    """
    coltype = _coltype_to_typeengine(coltype)
    if isinstance(coltype, sqltypes.Text):
        return True
    if isinstance(coltype, sqltypes.LargeBinary):
        return True
    return False


# =============================================================================
# Hack in new type
# =============================================================================

def hack_in_mssql_xml_type():
    r"""
    Modifies SQLAlchemy's type map for Microsoft SQL Server to support XML.
    
    SQLAlchemy does not support the XML type in SQL Server (mssql).
    Upon reflection, we get:
    
    .. code-block:: none
    
       sqlalchemy\dialects\mssql\base.py:1921: SAWarning: Did not recognize type 'xml' of column '...'

    We will convert anything of type ``XML`` into type ``TEXT``.

    """  # noqa
    log.debug("Adding type 'xml' to SQLAlchemy reflection for dialect 'mssql'")
    mssql.base.ischema_names['xml'] = mssql.base.TEXT
    # https://stackoverflow.com/questions/32917867/sqlalchemy-making-schema-reflection-find-use-a-custom-type-for-all-instances  # noqa

    # print(repr(mssql.base.ischema_names.keys()))
    # print(repr(mssql.base.ischema_names))


# =============================================================================
# Check column definition equality
# =============================================================================

def column_types_equal(a_coltype: TypeEngine, b_coltype: TypeEngine) -> bool:
    """
    Checks that two SQLAlchemy column types are equal (by comparing ``str()``
    versions of them).
    
    See https://stackoverflow.com/questions/34787794/sqlalchemy-column-type-comparison.
    
    IMPERFECT. 
    """  # noqa
    return str(a_coltype) == str(b_coltype)


def columns_equal(a: Column, b: Column) -> bool:
    """
    Are two SQLAlchemy columns are equal? Checks based on:

    - column ``name``
    - column ``type`` (see :func:`column_types_equal`)
    - ``nullable``
    """
    return (
        a.name == b.name and
        column_types_equal(a.type, b.type) and
        a.nullable == b.nullable
    )


def column_lists_equal(a: List[Column], b: List[Column]) -> bool:
    """
    Are all columns in list ``a`` equal to their counterparts in list ``b``,
    as per :func:`columns_equal`?
    """
    n = len(a)
    if len(b) != n:
        return False
    for i in range(n):
        if not columns_equal(a[i], b[i]):
            log.debug("Mismatch: {!r} != {!r}", a[i], b[i])
            return False
    return True


def indexes_equal(a: Index, b: Index) -> bool:
    """
    Are two indexes equal? Checks by comparing ``str()`` versions of them.
    (AM UNSURE IF THIS IS ENOUGH.)
    """
    return str(a) == str(b)


def index_lists_equal(a: List[Index], b: List[Index]) -> bool:
    """
    Are all indexes in list ``a`` equal to their counterparts in list ``b``,
    as per :func:`indexes_equal`?
    """
    n = len(a)
    if len(b) != n:
        return False
    for i in range(n):
        if not indexes_equal(a[i], b[i]):
            log.debug("Mismatch: {!r} != {!r}", a[i], b[i])
            return False
    return True


# =============================================================================
# Tests
# =============================================================================

def test_assert(x, y) -> None:
    try:
        assert x == y
    except AssertionError:
        print(f"{x!r} should have been {y!r}")
        raise


def unit_tests() -> None:
    from sqlalchemy.dialects.mssql.base import MSDialect
    from sqlalchemy.dialects.mysql.base import MySQLDialect
    d_mssql = MSDialect()
    d_mysql = MySQLDialect()
    col1 = Column('hello', BigInteger, nullable=True)
    col2 = Column('world', BigInteger,
                  autoincrement=True)  # does NOT generate IDENTITY
    col3 = make_bigint_autoincrement_column('you', d_mssql)
    metadata = MetaData()
    t = Table('mytable', metadata)
    t.append_column(col1)
    t.append_column(col2)
    t.append_column(col3)

    print("Checking Column -> DDL: SQL Server (mssql)")
    test_assert(column_creation_ddl(col1, d_mssql), "hello BIGINT NULL")
    test_assert(column_creation_ddl(col2, d_mssql), "world BIGINT NULL")
    test_assert(column_creation_ddl(col3, d_mssql),
                "you BIGINT NOT NULL IDENTITY(1,1)")

    print("Checking Column -> DDL: MySQL (mysql)")
    test_assert(column_creation_ddl(col1, d_mysql), "hello BIGINT")
    test_assert(column_creation_ddl(col2, d_mysql), "world BIGINT")
    # not col3; unsupported

    print("Checking SQL type -> SQL Alchemy type")
    to_check = [
        # mssql
        ("BIGINT", d_mssql),
        ("NVARCHAR(32)", d_mssql),
        ("NVARCHAR(MAX)", d_mssql),
        ('NVARCHAR(160) COLLATE "Latin1_General_CI_AS"', d_mssql),
        # mysql
        ("BIGINT", d_mssql),
        ("LONGTEXT", d_mysql),
        ("ENUM('red','green','blue')", d_mysql),
    ]
    for coltype, dialect in to_check:
        print(f"... {coltype!r} -> dialect {dialect.name!r} -> "
              f"{get_sqla_coltype_from_dialect_str(coltype, dialect)!r}")


if __name__ == '__main__':
    unit_tests()



#!/usr/bin/env python
# cardinal_pythonlib/django/fields/jsonclassfield.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Django field class implementing storage of arbitrary Python objects in a
database, so long as they are serializable to/from JSON.**

- We were using ``django-picklefield`` and ``PickledObjectField``.

- However, this fails in a nasty way if you add new attributes to a class
  that has been pickled, and anyway pickle is insecure (as it trusts its
  input).
  
- JSON is better.
  https://www.benfrederickson.com/dont-pickle-your-data/

- JSON fields in Django:
  https://djangopackages.org/grids/g/json-fields/

- https://paltman.com/how-to-store-arbitrary-data-in-a-django-model/

- Native Django JSONField requires PostgreSQL, and is not part of the core set
  of fields:

  https://docs.djangoproject.com/en/1.10/ref/contrib/postgres/fields/#django.contrib.postgres.fields.JSONField
  https://docs.djangoproject.com/en/1.10/ref/models/fields/

- https://stackoverflow.com/questions/6578986/how-to-convert-json-data-into-a-python-object
- https://stackoverflow.com/questions/31235771/is-parsing-a-json-naively-into-a-python-class-or-struct-secure
- https://stackoverflow.com/questions/16405969/how-to-change-json-encoding-behaviour-for-serializable-python-object/16406798#16406798
- https://stackoverflow.com/questions/3768895/how-to-make-a-class-json-serializable

e.g.:

.. code-block:: python
    
    import inspect
    import json
    from typing import Any, Dict, Union
    
    class Thing(object):
        def __init__(self, a: int = 1, b: str = ''):
            self.a = a
            self.b = b
        def __repr__(self) -> str:
            return "<Thing(a={}, b={}) at {}>".format(
                repr(self.a), repr(self.b), hex(id(self)))
    
    
    MY_JSON_TYPES = {
        'Thing': Thing,
    }
    TYPE_LABEL = '__type__'
    
    class MyEncoder(json.JSONEncoder):
        def default(self, obj: Any) -> Any:
            typename = type(obj).__name__
            if typename in MY_JSON_TYPES.keys():
                d = obj.__dict__
                d[TYPE_LABEL] = typename
                return d
            return super().default(obj)
    
    
    class MyDecoder(json.JSONDecoder):  # INADEQUATE for nested things
        def decode(self, s: str) -> Any:
            o = super().decode(s)
            if isinstance(o, dict):
                typename = o.get(TYPE_LABEL, '')
                if typename and typename in MY_JSON_TYPES:
                    classtype = MY_JSON_TYPES[typename]
                    o.pop(TYPE_LABEL)
                    return classtype(**o)
            return o
    
    
    def my_decoder_hook(d: Dict) -> Any:
        if TYPE_LABEL in d:
            typename = d.get(TYPE_LABEL, '')
            if typename and typename in MY_JSON_TYPES:
                classtype = MY_JSON_TYPES[typename]
                d.pop(TYPE_LABEL)
                return classtype(**d)
        return d
    
    
    x = Thing(a=5, b="hello")
    y = [1, x, 2]
    
    # Encoding:
    j = MyEncoder().encode(x)  # OK
    j2 = json.dumps(x, cls=MyEncoder)  # OK; same result
    
    k = MyEncoder().encode(y)  # OK
    k2 = json.dumps(y, cls=MyEncoder)  # OK; same result
    
    # Decoding
    x2 = MyDecoder().decode(j)  # OK, but simple structure
    y2 = MyDecoder().decode(k)  # FAILS
    y3 = json.JSONDecoder(object_hook=my_decoder_hook).decode(k)  # SUCCEEDS
    
    print(repr(x))
    print(repr(x2))

"""  # noqa

# noinspection PyUnresolvedReferences
from django.core.exceptions import ValidationError
# noinspection PyUnresolvedReferences
from django.db.models import TextField

from cardinal_pythonlib.json.serialize import json_decode, json_encode


# =============================================================================
# Django field
# - To use a class with this, the class must be registered with
#   register_class_for_json() above. Register the class immediately after
#   defining it.
# =============================================================================

class JsonClassField(TextField):
    """
    Django field that serializes Python objects into JSON.
    """
    # https://docs.djangoproject.com/en/1.10/howto/custom-model-fields/
    description = "Python objects serialized into JSON"

    # No need to implement __init__()
    # No need to implement deconstruct()
    # No need to implement db_type()

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def from_db_value(self, value, expression, connection):
        """
        "Called in all circumstances when the data is loaded from the
        database, including in aggregates and values() calls."
        """
        if value is None:
            return value
        return json_decode(value)

    def to_python(self, value):
        """
        "Called during deserialization and during the clean() method used
        from forms.... [s]hould deal gracefully with... (*) an instance of
        the correct type; (*) a string; (*) None (if the field allows
        null=True)."

        "For ``to_python()``, if anything goes wrong during value conversion,
        you should raise a ``ValidationError`` exception."
        """
        if value is None:
            return value
        if not isinstance(value, str):
            return value
        try:
            return json_decode(value)
        except Exception as err:
            raise ValidationError(repr(err))

    def get_prep_value(self, value):
        """
        Converse of ``to_python()``. Converts Python objects back to query
        values.
        """
        return json_encode(value)



#!/usr/bin/env python
# cardinal_pythonlib/platformfunc.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Support for platform-specific problems.**
"""

from collections import OrderedDict
import itertools
from pprint import pformat
import subprocess
import sys
from typing import Any, Dict, Generator, Iterator, List, Tuple, Union

from cardinal_pythonlib.fileops import require_executable
from cardinal_pythonlib.logs import get_brace_style_log_with_null_handler

log = get_brace_style_log_with_null_handler(__name__)


# =============================================================================
# Fix UTF-8 output problems on Windows
# =============================================================================
# https://stackoverflow.com/questions/5419

def fix_windows_utf8_output() -> None:
    # Python 3 only now, so nothing to do
    return

    # if six.PY3:
    #     return
    # reload_module(sys)
    # # noinspection PyUnresolvedReferences
    # sys.setdefaultencoding('utf-8')
    # # print sys.getdefaultencoding()
    #
    # if sys.platform == 'win32':
    #     try:
    #         import win32console
    #     except ImportError:
    #         win32console = None
    #         print(
    #             "Python Win32 Extensions module is required.\n "
    #             "You can download it from "
    #             "https://sourceforge.net/projects/pywin32/ "
    #             "(x86 and x64 builds are available)\n")
    #         exit(-1)
    #     # win32console implementation  of SetConsoleCP does not return a value
    #     # CP_UTF8 = 65001
    #     win32console.SetConsoleCP(65001)
    #     if win32console.GetConsoleCP() != 65001:
    #         raise RuntimeError("Cannot set console codepage to 65001 (UTF-8)")
    #     win32console.SetConsoleOutputCP(65001)
    #     if win32console.GetConsoleOutputCP() != 65001:
    #         raise RuntimeError("Cannot set console output codepage to 65001 "
    #                            "(UTF-8)")
    #
    # sys.stdout = codecs.getwriter('utf8')(sys.stdout)
    # sys.stderr = codecs.getwriter('utf8')(sys.stderr)
    # # CHECK: does that modify the "global" sys.stdout?
    # # You can't use "global sys.stdout"; that raises an error


def test_windows_utf8_output() -> None:
    """
    Print a short string with unusual Unicode characters.
    """
    print(u"This is an Е乂αmp١ȅ testing Unicode support using Arabic, Latin, "
          u"Cyrillic, Greek, Hebrew and CJK code points.\n")


if __name__ == '__main__':
    fix_windows_utf8_output()
    test_windows_utf8_output()


# =============================================================================
# Check package presence on Debian
# =============================================================================

DPKG_QUERY = "dpkg-query"


def are_debian_packages_installed(packages: List[str]) -> Dict[str, bool]:
    """
    Check which of a list of Debian packages are installed, via ``dpkg-query``.

    Args:
        packages: list of Debian package names

    Returns:
        dict: mapping from package name to boolean ("present?")

    """
    assert len(packages) >= 1
    require_executable(DPKG_QUERY)
    args = [
        DPKG_QUERY,
        "-W",  # --show
        # "-f='${Package} ${Status} ${Version}\n'",
        "-f=${Package} ${Status}\n",  # --showformat
    ] + packages
    completed_process = subprocess.run(args,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE,
                                       check=False)
    encoding = sys.getdefaultencoding()
    stdout = completed_process.stdout.decode(encoding)
    stderr = completed_process.stderr.decode(encoding)
    present = OrderedDict()
    for line in stdout.split("\n"):
        if line:  # e.g. "autoconf install ok installed"
            words = line.split()
            assert len(words) >= 2
            package = words[0]
            present[package] = "installed" in words[1:]
    for line in stderr.split("\n"):
        if line:  # e.g. "dpkg-query: no packages found matching XXX"
            words = line.split()
            assert len(words) >= 2
            package = words[-1]
            present[package] = False
    log.debug("Debian package presence: {}", present)
    return present


def require_debian_packages(packages: List[str]) -> None:
    """
    Ensure specific packages are installed under Debian.

    Args:
        packages: list of packages

    Raises:
        ValueError: if any are missing

    """
    present = are_debian_packages_installed(packages)
    missing_packages = [k for k, v in present.items() if not v]
    if missing_packages:
        missing_packages.sort()
        msg = (
            f"Debian packages are missing, as follows. Suggest:\n\n"
            f"sudo apt install {' '.join(missing_packages)}"
        )
        log.critical(msg)
        raise ValueError(msg)


# =============================================================================
# Get the environment from a subprocess in Windows
# =============================================================================

def validate_pair(ob: Any) -> bool:
    """
    Does the object have length 2?
    """
    try:
        if len(ob) != 2:
            log.warning("Unexpected result: {!r}", ob)
            raise ValueError()
    except ValueError:
        return False
    return True


def consume(iterator: Iterator[Any]) -> None:
    """
    Consume all remaining values of an iterator.

    A reminder: iterable versus iterator:
    https://anandology.com/python-practice-book/iterators.html.
    """
    try:
        while True:
            next(iterator)
    except StopIteration:
        pass


def windows_get_environment_from_batch_command(
        env_cmd: Union[str, List[str]],
        initial_env: Dict[str, str] = None) -> Dict[str, str]:
    """
    Take a command (either a single command or list of arguments) and return
    the environment created after running that command. Note that the command
    must be a batch (``.bat``) file or ``.cmd`` file, or the changes to the
    environment will not be captured.

    If ``initial_env`` is supplied, it is used as the initial environment
    passed to the child process. (Otherwise, this process's ``os.environ()``
    will be used by default.)

    From https://stackoverflow.com/questions/1214496/how-to-get-environment-from-a-subprocess-in-python,
    with decoding bug fixed for Python 3.

    PURPOSE: under Windows, ``VCVARSALL.BAT`` sets up a lot of environment
    variables to compile for a specific target architecture. We want to be able
    to read them, not to replicate its work.

    METHOD: create a composite command that executes the specified command,
    then echoes an unusual string tag, then prints the environment via ``SET``;
    capture the output, work out what came from ``SET``.

    Args:
        env_cmd: command, or list of command arguments
        initial_env: optional dictionary containing initial environment

    Returns:
        dict: environment created after running the command
    """  # noqa
    if not isinstance(env_cmd, (list, tuple)):
        env_cmd = [env_cmd]
    # construct the command that will alter the environment
    env_cmd = subprocess.list2cmdline(env_cmd)
    # create a tag so we can tell in the output when the proc is done
    tag = '+/!+/!+/! Finished command to set/print env +/!+/!+/!'  # RNC
    # construct a cmd.exe command to do accomplish this
    cmd = f'cmd.exe /s /c "{env_cmd} && echo "{tag}" && set"'
    # launch the process
    log.info("Fetching environment using command: {}", env_cmd)
    log.debug("Full command: {}", cmd)
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=initial_env)
    # parse the output sent to stdout
    encoding = sys.getdefaultencoding()

    def gen_lines() -> Generator[str, None, None]:  # RNC: fix decode problem
        for line in proc.stdout:
            yield line.decode(encoding)

    # define a way to handle each KEY=VALUE line
    def handle_line(line: str) -> Tuple[str, str]:  # RNC: as function
        # noinspection PyTypeChecker
        parts = line.rstrip().split('=', 1)
        # split("=", 1) means "split at '=' and do at most 1 split"
        if len(parts) < 2:
            return parts[0], ""
        return parts[0], parts[1]

    lines = gen_lines()  # RNC
    # consume whatever output occurs until the tag is reached
    consume(itertools.takewhile(lambda l: tag not in l, lines))
    # ... RNC: note that itertools.takewhile() generates values not matching
    #     the condition, but then consumes the condition value itself. So the
    #     tag's already gone. Example:
    #
    #   def gen():
    #       mylist = [1, 2, 3, 4, 5]
    #       for x in mylist:
    #           yield x
    #
    #   g = gen()
    #   list(itertools.takewhile(lambda x: x != 3, g))  # [1, 2]
    #   next(g)  # 4, not 3
    #
    # parse key/values into pairs
    pairs = map(handle_line, lines)
    # make sure the pairs are valid (this also eliminates the tag)
    valid_pairs = filter(validate_pair, pairs)
    # construct a dictionary of the pairs
    result = dict(valid_pairs)  # consumes generator
    # let the process finish
    proc.communicate()
    log.debug("Fetched environment:\n" + pformat(result))
    return result


# =============================================================================
# Check for a special environment danger (vulnerability) in Windows
# =============================================================================

def contains_unquoted_target(x: str,
                             quote: str = '"', target: str = '&') -> bool:
    """
    Checks if ``target`` exists in ``x`` outside quotes (as defined by
    ``quote``). Principal use: from
    :func:`contains_unquoted_ampersand_dangerous_to_windows`.
    """
    in_quote = False
    for c in x:
        if c == quote:
            in_quote = not in_quote
        elif c == target:
            if not in_quote:
                return True
    return False


def contains_unquoted_ampersand_dangerous_to_windows(x: str) -> bool:
    """
    Under Windows, if an ampersand is in a path and is not quoted, it'll break
    lots of things.
    See https://stackoverflow.com/questions/34124636.
    Simple example:

    .. code-block:: bat

        set RUBBISH=a & b           # 'b' is not recognizable as a... command
        set RUBBISH='a & b'         # 'b'' is not recognizable as a... command
        set RUBBISH="a & b"         # OK

    ... and you get similar knock-on effects, e.g. if you set RUBBISH using the
    Control Panel to the literal

    .. code-block:: bat

        a & dir

    and then do

    .. code-block:: bat

        echo %RUBBISH%

    it will (1) print "a" and then (2) print a directory listing as it RUNS
    "dir"! That's pretty dreadful.

    See also
        https://www.thesecurityfactory.be/command-injection-windows.html

    Anyway, this is a sanity check for that sort of thing.
    """
    return contains_unquoted_target(x, quote='"', target='&')



#!/usr/bin/env python
# cardinal_pythonlib/ui.py

"""
===============================================================================

    Original code copyright (C) 2009-2021 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of cardinal_pythonlib.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        https://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

===============================================================================

**Support functions for user interaction -- command line only.**

Does not need tkinter.

"""

import getpass
from typing import Optional


def ask_user(prompt: str, default: str = None) -> Optional[str]:
    """
    Prompts the user, with a default. Returns user input from ``stdin``.
    """
    if default is None:
        prompt += ": "
    else:
        prompt += " [" + default + "]: "
    result = input(prompt)
    return result if len(result) > 0 else default


def ask_user_password(prompt: str) -> str:
    """
    Read a password from the console.
    """
    return getpass.getpass(prompt + ": ")
